package ims.oracle;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import ims.oracle.Configuration.TR;
import ims.oracle.DeterministicArcStandardSwapOldOracle.InOrderOrder;
import ims.trans.Options.TransitionSystem;
import ims.util.DisjointSetZeroToN;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class DeterministicArcStandardOptimalSwapOracle extends AbstractStaticOracle{
	
	/*
	 * Building the lattice breadth-first style and then stopping at the first terminal state doesn't work.
	 * We have to do depth-first style search, and store the signatures for every state.
	 * 
	 * So we need a map that maps signatures -> transition sequences. When we reach a previously seen state we have two options:
	 * 1) If the transition sequence that took us there this time is of equal or longer length than the state already there, stop and continue with the agenda
	 * 2) If the transition sequence that took us there this time is shorter than the previous one we have to continue from here (as this probably is a shorter path then)
	 * 
	 */
	
	// the other option here would be to do *no caching* at all, and just always go to the end. Requires more computational power but is less memory intensive.

	protected DeterministicArcStandardOptimalSwapOracle(TR[] myTRs) {
		super(myTRs);
	}


	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels,boolean labeled) {

		TObjectIntMap<String> seen=new TObjectIntHashMap<String>();
		List<MinimalConfigurationWTRSig> agenda=new ArrayList<MinimalConfigurationWTRSig>();
		InOrderOrder inOrder=new InOrderOrder(InOrderOrder.getSequence(heads));
		DisjointSetZeroToN mpcs=DeterministicArcStandardSwapNewOracle.getMPCs(heads, new Configuration(heads));
		MinimalConfigurationWTRSig start=new MinimalConfigurationWTRSig(heads);		
		MinimalConfigurationWTRSig optimal=null;
		agenda.add(start);
		while(!agenda.isEmpty()){
//			System.err.println("Agenda size: "+agenda.size()+", Seen size: "+seen.size()+", Opt len: "+(optimal==null?"-1":optimal.getTransitionSeqSignature().length()-1));
			MinimalConfigurationWTRSig cur=agenda.remove(agenda.size()-1); //next item from the agenda
			while(!cur.isTerminal()){
				String curTrSig=cur.getTransitionSeqSignature();
				String curStSig=cur.getStackBufSig();
				if(optimal!=null && optimal.getTransitionSeqSignature().length()<=curTrSig.length()) //Can break if we've already made more transitions than an optimal
					break;
				if(seen.containsKey(curStSig) && seen.get(curStSig)<=curTrSig.length()) //Can break if we've seen this state before with a shorter or equally long tr seq
					break;
				seen.put(curStSig, curTrSig.length());
				List<TR> actions=getActions(heads, cur, inOrder, mpcs);
				if(actions.size()>1){ 
					//More than one possible action, then we have to fork save all but one in the agenda for later
					//Since we only allow for the {SH,SW} ambiguity, then we go for delaying the swap
					MinimalConfigurationWTRSig fork=cur.deepCopy();
					fork.doSwap();
					agenda.add(fork);
					cur.doShift();
				} else {
					cur.applyTR(actions.get(0)); //If there's only one possible transition then we don't need to clone or anything, can just go on.
				}
			}
			if(cur.isTerminal()){
				if(optimal==null || optimal.getTransitionSeqSignature().length()>cur.getTransitionSeqSignature().length())
					optimal=cur;
			}
		}

		return getReturnSeq(optimal,labeled,heads,labels);
	}


	private List<String> getReturnSeq(IConfiguration n, boolean labeled,int[] heads,String[] labels) {
		String trSig=n.getTransitionSeqSignature();
		List<String> r=new ArrayList<String>();
		Configuration c=new Configuration(heads);
		for(int i=1;i<trSig.length();++i){
			TR tr=TR.fromC(trSig.charAt(i));
			String t;
			switch(tr){
			case SH:  t=tr.toString(); break; 
			case SW:  t=tr.toString(); break;
			case LA:  t=tr.toString()+(labeled?"_"+labels[c.peekStack(1)]:""); break;
			case RAS: t=tr.toString()+(labeled?"_"+labels[c.peekStack(0)]:""); break;
			default: throw new Error("you are wrong here");
			}
			r.add(t);
			c.applyTR(tr);
		}
		return r;
	}

	private List<TR> getActions(int[] heads, IConfiguration c, InOrderOrder inOrder,DisjointSetZeroToN mpcs) {
		
		if(DeterministicArcStandardOracle.canLeftArc(heads, c))
			return Collections.singletonList(TR.LA);
		
		if(DeterministicArcStandardOracle.canRightArc(heads, c))
			return Collections.singletonList(TR.RAS);
		
//		boolean canSwap=DeterministicArcStandardSwapOldOracle.canSwap(heads, c, inOrder);
		boolean canSwap=DeterministicArcStandardSwapNewOracle.canSwap(heads, c, inOrder, mpcs);
		if(canSwap){ //can we shift too?
			boolean canShift=BruteForceShiftOKCheck.canShift(c, inOrder, heads);
			if(canShift){
				return Arrays.asList(TR.SH,TR.SW);
			} else {
				return Collections.singletonList(TR.SW);
			}
		} else {
			return Collections.singletonList(TR.SH);
		}
	}

	@Override
	public boolean canDoNonProj() {
		return true;
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, int[] heads) {
		int i=conf.getTotalTransitions(); //correct transition will be at this position in the cachedCanonicalSeq
		if(i>=cachedCanonicalSeq.size())
			return false;
		return cachedCanonicalSeq.get(i).equals(tr.toString());
	}

	private List<String> cachedCanonicalSeq;
	
	@Override
	public void initSentence(int[] heads){
		this.cachedCanonicalSeq=getCanonicalSequence(heads,null,false);
	}
	
	@Override
	public TransitionSystem getTS() {
		return TransitionSystem.ArcSOptimalSwap;
	}
}
