package ims.oracle;

import java.util.Arrays;

import ims.data.Buffer;
import ims.data.ISentence;
import ims.data.Stack;
import ims.oracle.Configuration.TR;
import ims.trans.State;

public class MinimalConfiguration implements IConfiguration{

	final Stack stack;
	final Buffer buffer;
	final int[] remainingDeps;
	final int[] predHeads;
	int totalTransitions;
//	String trSig;
	
	protected MinimalConfiguration(Stack stack,Buffer buffer,int[] remainingDeps,int totalTransitions,int[] predHeads){
		this.stack=stack;
		this.buffer=buffer;
		this.remainingDeps=remainingDeps;
		this.predHeads=predHeads;
		this.totalTransitions=totalTransitions;
	}
	
	public MinimalConfiguration(int[] heads) {
		this.stack=new Stack();
		this.buffer=new Buffer(heads.length-1);
		this.remainingDeps=new int[heads.length];
		this.totalTransitions=0;
		this.predHeads=new int[heads.length];
		this.predHeads[0]=ISentence.ROOT_HEAD;
		//Then fill the remainigDeps
		for(int i=1;i<heads.length;++i)
			this.remainingDeps[heads[i]]++;
	}

	@Override
	public int bufferSize() {
		return buffer.size();
	}

	@Override
	public int stackSize() {
		return stack.size();
	}

	@Override
	public boolean hasMinTwoStackElements() {
		return stack.size()>1;
	}

	@Override
	public int peekBuffer(int d) {
		return buffer.get(d);
	}

	@Override
	public int peekStack(int d) {
		return stack.peek(d);
	}

	@Override
	public boolean hasHead(int dep) {
		return getHead(dep)>=0;
	}

	@Override
	public boolean hasAllDeps(int head) {
		return remainingDeps[head]==0;
	}

	@Override
	public boolean hasAllLeftDeps(int head) {
		throw new UnsupportedOperationException("not implemented");
	}

	@Override
	public boolean hasAllRightDeps(int head) {
		throw new UnsupportedOperationException("not implemented");
	}

	@Override
	public TokenState getTokenState(int i) {
		throw new UnsupportedOperationException("not implemented");
	}
	
	public void doShift(){
		stack.add(buffer.removeFront());
//		trSig+=TR.SH.k;
		
		++totalTransitions;
	}
	
	public void doLA(){
		int s0=stack.pop();
		int s1=stack.pop(); //s1 goes, we don't need it any longer
		stack.add(s0);
		--remainingDeps[s0];
//		trSig+=TR.LA.k;
		
		int head=s0;
		int dep=s1;
		predHeads[dep]=head;
		
		++totalTransitions;
	}
	
	public void doRA(){
		int s0=stack.pop(); //s0 goes, we don't need it any longer
		int s1=stack.peek(0);
		--remainingDeps[s1];
//		trSig+=TR.RAS.k;
		
		int dep=s0;
		int head=s1;
		predHeads[dep]=head;

		++totalTransitions;
	}
	
	public void doSwap(){
		int s0=stack.pop();
		int s1=stack.pop();
		stack.add(s0);
		buffer.pushBack(s1);
//		trSig+=TR.SW.k;
		
		++totalTransitions;
	}

	@Override
	public int sentenceSize() {
		return predHeads.length;
	}

	@Override
	public int getBufferNoPushedBack() {
		return buffer.getBufferNoPushedBack();
	}

	@Override
	public String getTransitionSeqSignature() {
		throw new UnsupportedOperationException("not supported");
	}

	public MinimalConfiguration deepCopy(){
		Stack s=stack.copy();
		Buffer b=buffer.copy();
		int[] rD=Arrays.copyOf(remainingDeps, remainingDeps.length);
		int[] pH=Arrays.copyOf(predHeads, predHeads.length);
		return new MinimalConfiguration(s,b,rD,totalTransitions,pH);
	}

	public void applyTR(TR tr) {
		switch(tr){
		case SH:  doShift(); break;
		case SW:  doSwap(); break;
		case LA:  doLA(); break;
		case RAS: doRA(); break;
		default: throw new Error("!");
		}
	}

	public boolean isTerminal() {
		return stack.size()==1 && buffer.size()==0;
	}

	@Override
	public int getHead(int d) {
		return predHeads[d];
	}
	
	public int[] getHeads(){
		return predHeads;
	}

	@Override
	public int getLbl(int i) {
		throw new UnsupportedOperationException("not implemented");
	}

	@Override
	public StringBuilder appendStackSig(StringBuilder sb) {
		return Configuration.appendStackSigLinear(sb, this);
	}

	@Override
	public String getULBLStateSig() {
		return State.getULBLSig(this);
	}

	@Override
	public String getLBLStateSig() {
		throw new UnsupportedOperationException("not implemented");
	}

	@Override
	public int getRemainingDeps(int head) {
		return remainingDeps[head];
	}
	
	public String getStackBufSig(){
		StringBuilder sb=new StringBuilder(stack.size()+buffer.size());
		for(int i=0;i<buffer.size();++i)
			sb.append((char) buffer.get(i));
		for(int i=stack.size()-1;i>=0;--i)
			sb.append((char) stack.peek(i));
		return sb.toString();
	}

	public static MinimalConfiguration getMinimalConfigurationCopyShiftedByOne(IConfiguration c,int length){
//		Stack stack=new Stack();
//		for(int d=c.stackSize()-2;d>=0;--d) //skip root, so do -2
//			stack.add(c.peekStack(d));
		Stack stack=c.getStackCopy();
		stack.add(c.peekBuffer(0));
		Buffer buffer=Buffer.copyOfBufferShiftedOne(c);
		int[] remainingDeps=new int[length];
		for(int i=0;i<length;++i)
			remainingDeps[i]=c.getRemainingDeps(i);
//		String trSig=c.getTransitionSeqSignature();
		
		int[] predHeads=new int[length];
		return new MinimalConfiguration(stack, buffer, remainingDeps, c.getTotalTransitions()+1, predHeads);
	}

	@Override
	public Stack getStackCopy() {
		return stack.copy();
	}

	@Override
	public int getTotalTransitions() {
		return totalTransitions;
	}
}
