package ims.perceptron;

import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.procedure.TIntIntProcedure;
import gnu.trove.procedure.TIntProcedure;
import ims.perceptron.Regularizer.RegularizerTypes;
import ims.trans.Options;

import java.util.Arrays;

public class ParametersFloat {

	public final float[] parameters;
	public final float[] total;  //For perceptron averaging
	public final float[] diag_G; //For AdaGrad
	public final float[] u;      //For l1-adagrad
	
	public final Regularizer reg;
	
	public ParametersFloat(int size,Regularizer reg){
		parameters=new float[size];
		total=new float[size];
		switch(reg.getType()){
		case AdaGradL1: 
		case AdaGrad:
			diag_G=new float[size];
			Arrays.fill(diag_G, 1.f);
			break;
		default:
			diag_G=null;
			break;
		}
		switch(reg.getType()){
		case AdaGradL1:
			u=new float[size];
			break;
		default:
			u=null;
			break;
		}
		this.reg=reg;
	}
	
	public void average(float avVal) {
		for(int j=0;j<total.length;++j)
			parameters[j]-=total[j]/avVal;
		Arrays.fill(total,Float.NEGATIVE_INFINITY);
	}
	public float[] averagedCopy(float avVal){
		float[] r=new float[parameters.length];
		for(int i=0;i<r.length;++i)
			r[i]=parameters[i]-total[i]/avVal;
		return r;
	}
	
	public int countNZ(){
		int notZero=0;
		for(int i=0;i<parameters.length;i++)
			if(parameters[i]!=0.f)
				notZero++; 
		return notZero;
	}
	
	public String getNonZeroWeights(boolean indicesOnly){
		StringBuilder sb=new StringBuilder();
		for(int i=0;i<parameters.length;i++)
			if (parameters[i]!=0.0F){
				sb.append(i);
				if(!indicesOnly) sb.append('=').append(parameters[i]);
				sb.append(", ");
			}
		return sb.toString();
	}
	
	public IUpdStruct getUpdStruct(double predScore,double actScore,double loss,double upd,TIntIntHashMap dv){
		if(reg.getType()==RegularizerTypes.AdaGrad){
			return new AdaGradUpdStruct(dv,upd);
		} else if(reg.getType()==RegularizerTypes.AdaGradL1){
			return new AdaGradL1UpdStruct(dv,upd);
		} else {
			return new PerceptronUpdStruct(predScore, actScore, loss, upd, dv);
		}
	}
	
	public static abstract class IUpdStruct {
		public abstract boolean apply(ParametersFloat parameters);
	}
	
	public class AdaGradL1UpdStruct extends AdaGradUpdStruct {

		public AdaGradL1UpdStruct(TIntIntHashMap dv, double upd) {
			super(dv, upd);
		}
		
		@Override
		public boolean apply(ParametersFloat parameters) {
			AdaGradL1UpdateProcedure up=new AdaGradL1UpdateProcedure(reg.getEta(),upd,reg.getLambda());
			dv.forEachEntry(up);
			return true;

		}
	}
	
	public class AdaGradUpdStruct extends IUpdStruct {
		final TIntIntHashMap dv;
		final double upd;
		
		public AdaGradUpdStruct(TIntIntHashMap dv,double upd){
			this.dv=dv;
			this.upd=upd;
		}
		
		@Override
		public boolean apply(ParametersFloat parameters) {
			AdaGradUpdateProcedure up=new AdaGradUpdateProcedure(reg.getEta(),upd);
			dv.forEachEntry(up);
			return true;
		}
		
	}
	
	public class PerceptronUpdStruct extends IUpdStruct {
		final TIntIntHashMap dv;
		final double num;
		final double upd;
		
		public PerceptronUpdStruct(double predScore,double actScore,double loss,double upd,TIntIntHashMap dv){
			if(predScore<actScore && !Options.AUG_LOSS)
				System.out.println("\nPred score less than act score!!: "+predScore +" < "+actScore+"  [THIS IS A SERIOUS PROBLEM AND YOU PROBABLY HAVE SEARCH ERRORS]");
			this.dv=dv;
			this.upd=upd;
			this.num=predScore-actScore+loss;
		}

		@Override
		public boolean apply(ParametersFloat parameters) {
			if(dv==null || dv.isEmpty())
				return false;
			float distTwoNormSquared=0;
			if(reg.needsDistNorm()){
				distTwoNormSquared=selfDotProduct(dv);
				if(distTwoNormSquared<0.00000001){
					System.out.println("\nRETURNING DUE TO TOO SMALL DIST NORM: "+distTwoNormSquared);
					return false;
				}
			}
			double tau=reg.getTau(num, distTwoNormSquared);
			if(tau<0.00000001){
				System.out.println("\nReturning due to too small tau: "+tau);
				return false;
			}
			update(tau,upd,dv);
			return true;
		}
	
		private void update(double tau,double upd,TIntIntHashMap m){
			PerceptronUpdateProcedure up=new PerceptronUpdateProcedure(tau,upd);
			m.forEachEntry(up);
		}
	}
	
	//This is based on the brief pdf by Chris Dyer that you find when you google AdaGrad
	//
	//One thing that's unclear in this pdf is how diagonal of the G matrix should be initialized.
	//Since nothing is said, I'd assume 0, but that would cause division by zero at the first update.
	//Hence I default it to one.
	//
	//I don't think AdaGrad needs to be averaged, but we keep the logic here so that we can try it
	class AdaGradUpdateProcedure implements TIntIntProcedure {
		final double eta;
		final double upd;
		
		AdaGradUpdateProcedure(double eta,double upd){
			this.eta=eta;
			this.upd=upd;
		}
		@Override
		public boolean execute(int arg0, int arg1) {
			double eta_i=eta/Math.sqrt(diag_G[arg0]); //Compute eta for this round
			diag_G[arg0]+=arg1*arg1;                  //Update running sum
			parameters[arg0]+=arg1*eta_i;             //Do update
			total[arg0]+=arg1*eta_i*upd;              //For averaging 
			return true;
		}
		
	}
	
	class AdaGradL1UpdateProcedure implements TIntIntProcedure {
		final double eta;
		final double upd;
		final double lambda;
		
		AdaGradL1UpdateProcedure(double eta,double upd,double lambda){
			this.eta=eta;
			this.upd=upd;
			this.lambda=lambda;
		}
		@Override
		public boolean execute(int arg0, int arg1) {
			final double nextW; //new weight for index arg0
			double x=Math.abs(u[arg0]/upd);
			if(x<lambda)
				nextW=0;
			else
				nextW=Math.signum(u[arg0])*eta*upd*(x-lambda)/Math.sqrt(diag_G[arg0]);
			
			parameters[arg0]=(float) nextW;
			
			diag_G[arg0]+=arg1*arg1;                  //Update running sum of squares
			total[arg0]+=nextW*upd;              //For averaging
			u[arg0]+=arg1;                            //Update running sum
			return true;
		}
		
	}

	
	class PerceptronUpdateProcedure implements TIntIntProcedure{
		final double tau;
		final double upd;
		
		PerceptronUpdateProcedure(double tau,double upd){
			this.tau=tau;
			this.upd=upd;
		}
		@Override
		public boolean execute(int arg0, int arg1) {
			parameters[arg0]+=arg1*tau;
			total[arg0]+=upd*arg1*tau;
			return true;
		}
	}

	
	public static int selfDotProduct(TIntIntHashMap m){
		SelfDotProductProcedure sdpp=new SelfDotProductProcedure();
		m.forEachValue(sdpp);
		return sdpp.sum;
	}
	static class SelfDotProductProcedure implements TIntProcedure {
		int sum=0;
		@Override
		public boolean execute(int arg0) {
			sum+=arg0*arg0;
			return true;
		}
	}

	public void update(IUpdStruct upd) {
		upd.apply(this);
	}
	
}
