package ims.perceptron;

import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.procedure.TIntIntProcedure;

public abstract class Regularizer {

	public abstract RegularizerTypes getType();
	public abstract double getTau(double numerator,double distTwoNormSquared);
	public abstract double getEta();
	public abstract double getLambda();
	public abstract String toString();
	public abstract boolean needsDistNorm();
	
	public static enum RegularizerTypes { PA0, PA1, PA2, Perceptron, AdaGrad, AdaGradL1 }
	
	public static Regularizer getRegularizer(RegularizerTypes t,double C,double eta,double lambda){
		switch(t){
		case PA0: return new PA0Regularizer();
		case PA1: return new PA1Regularizer(C);
		case PA2: return new PA2Regularizer(C);
		case Perceptron: return new RegularPerceptron();
		case AdaGrad:    return new AdaGrad(eta);
		case AdaGradL1:  return new AdaGradL1(eta,lambda);
		}
		throw new Error("not implemented");
	}

	/*
	 * This is the basic from Huang et al. (2012). Should probably be moved into the subclasses and implemented appropriately for all the other regularizers.
	 */
	public double getViolation(float[] params,TIntIntHashMap distVector){
		ScoreDistVectorProcedure sdvp=new ScoreDistVectorProcedure(params);
		distVector.forEachEntry(sdvp);
		return sdvp.sum;
	}
	
	
	//The two should be equivalent, only difference might be floatingpoint arithmetics.
	static final boolean USE_OLD_VIOLATION_COMPUTATION=false;
	
	public double getViolation(float[] params,IFV<FV3> goldFV,IFV<FV3> predFV){
		if(USE_OLD_VIOLATION_COMPUTATION){
			TIntIntHashMap distVector=goldFV.getDistVector(predFV);
			return getViolation(params,distVector);
		} else {
			double vio=predFV.getScore(params) - goldFV.getScore(params);
			return vio;
		}
	}
	
	static class ScoreDistVectorProcedure implements TIntIntProcedure {
		final float[] params;
		double sum=0.d;
		
		ScoreDistVectorProcedure(float[] params){
			this.params=params;
		}
		@Override
		public boolean execute(int arg0, int arg1) {
			sum-=arg1*params[arg0];
			return true;
		}
	}
	
	public static class AdaGradL1 extends Regularizer {
		public final double eta;
		public final double lambda;
		public AdaGradL1(double eta, double lambda){
			this.eta=eta;
			this.lambda=lambda;
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.AdaGradL1;
		}
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			throw new UnsupportedOperationException("there is no tau in adagrad");
		}
		@Override
		public double getEta() {
			return eta;
		}
		@Override
		public String toString() {
			return "AdaGradL1, eta="+eta+", lambda="+lambda;
		}
		@Override
		public double getLambda() {
			return lambda;
		}
		@Override
		public boolean needsDistNorm() {
			return false;
		}
	}
	
	public static class AdaGrad extends Regularizer {
		public final double eta;
		
		public AdaGrad(double eta){
			this.eta=eta;
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.AdaGrad;
		}
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			throw new UnsupportedOperationException("there is no tau in adagrad");
		}
		@Override
		public String toString() {
			return "AdaGrad, eta="+eta;
		}
		@Override
		public double getEta() {
			return eta;
		}
		@Override
		public double getLambda() {
			throw new UnsupportedOperationException("there is no lambda in vanilla adagrad");
		}
		@Override
		public boolean needsDistNorm() {
			return false;
		}
	}
	
	/*
	 * Regular perceptron update
	 */
	public static class RegularPerceptron extends Regularizer {
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			return 1.0f;
		}
		@Override
		public String toString() {
			return "Perceptron";
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.Perceptron;
		}
		@Override
		public double getEta() {
			throw new UnsupportedOperationException("there is no eta in "+getType());
		}
		@Override
		public double getLambda() {
			throw new UnsupportedOperationException("there is no lambda in perceptron");
		}
		@Override
		public boolean needsDistNorm() {
			return false;
		}
	}
	
	/*
	 * Loss functions as per Crammer et al 2006.
	 */
	public static class PA0Regularizer extends Regularizer {
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			return numerator/distTwoNormSquared;
		}
		@Override
		public String toString() {
			return "PA0";
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.PA0;
		}
		@Override
		public double getEta() {
			throw new UnsupportedOperationException("there is no eta in "+getType());
		}
		@Override
		public double getLambda() {
			throw new UnsupportedOperationException("there is no lambda in perceptron");
		}
		@Override
		public boolean needsDistNorm() {
			return true;
		}
	}
	public static class PA1Regularizer extends Regularizer {
		private final double C;
		public PA1Regularizer(double C){
			this.C=C;
		}
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			return Math.min(C, numerator/distTwoNormSquared);
		}
		@Override
		public String toString() {
			return "PA1, C="+C;
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.PA1;
		}
		@Override
		public double getEta() {
			throw new UnsupportedOperationException("there is no eta in "+getType());
		}
		@Override
		public double getLambda() {
			throw new UnsupportedOperationException("there is no lambda in perceptron");
		}
		@Override
		public boolean needsDistNorm() {
			return true;
		}
	}
	public static class PA2Regularizer extends Regularizer {
		private final double Cdouble;
		public PA2Regularizer(double C){
			this.Cdouble=2*C;
		}
		@Override
		public double getTau(double numerator, double distTwoNormSquared) {
			return numerator/(distTwoNormSquared+Cdouble);
		}
		@Override
		public String toString() {
			return "PA2, C="+(Cdouble/2.0d);
		}
		@Override
		public RegularizerTypes getType() {
			return RegularizerTypes.PA2;
		}
		@Override
		public double getEta() {
			throw new UnsupportedOperationException("there is no eta in "+getType());
		}
		@Override
		public double getLambda() {
			throw new UnsupportedOperationException("there is no lambda in perceptron");
		}
		@Override
		public boolean needsDistNorm() {
			return true;
		}
	}
}
