package ims.trans;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.IOracle;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.ParametersFloat.IUpdStruct;
import ims.trans.Decoder.TerminalState;
import ims.trans.Options.ViolationPolicy;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.util.ArrayFunctions;
import ims.util.DBO;
import ims.util.IntPair;
import ims.util.MinHeap;
import ims.util.Pair;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;

public abstract class AbstractBeamTrainer extends AbstractTrainer {

	public static final boolean DO_MEASURE_FORK_POINTS=true;
	
	protected final int goldBeam;
	
	public AbstractBeamTrainer(Options options, SymbolTable symTab,	IOracle oracle, ITransitionFilter transitionFilter,	List<Instance> instances, FeatureExtractor featureExtractor, Long2IntInterface l2i, ParametersFloat parameters, Decoder decoder, LossFunction lossFunction, InstanceMapping instanceMapping,int goldBeam) {
		super(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping);
		this.goldBeam=goldBeam;
	}
	
	void recoverFeaturesAndFillVector(IFV<?> fv, State finalState,Instance inst,boolean clearFv) {
		if(clearFv)
			fv.clear();
		ITransition[] transitions=State.getTransitionSequence(finalState);
		
		//Now move forward and extract the features again, and add them to the fv.
		//s is now start state.
		State s=State.getInitialState(inst.head.length,options.labeled,null);
//		long[] sink=new long[FeatureExtractor.SINK_MIN_SIZE];
		long[] sink=featureExtractor.sink;
		for(ITransition tr:transitions){
			int f=featureExtractor.extractFeatures(s, sink, inst);
			for(int j=0;j<f;++j){
				sink[j]<<=featureExtractor.bA;
				sink[j]|=tr.getId();
				fv.add(l2i.l2i(sink[j]));
			}
			s=tr.apply(s,0); //We don't care about the score here anyway, so we make it 0
		}
	}

	protected double doUpdate(int itemCount, IFV<FV3> goldFV, IFV<FV3> predFV,Instance inst, State bestPrediction, State correctState) {
		double goldScore=correctState.score();
		double predScore=bestPrediction.score();

		recoverFeaturesAndFillVector(goldFV,correctState,inst,true);
		recoverFeaturesAndFillVector(predFV,bestPrediction,inst,true);
		double loss=lossFunction.getLoss(correctState,bestPrediction,inst.form.length); //This is oLoss

		if(Options.AUG_LOSS){
			//Since we do inexact search, the augmented loss transitions can have side effects that incur additional loss in absolute terms. <--- Actually, I guess the reason isn't the inexact search but more a consequence of how certain arcs are ruled out by certain other actions... Or is it because the loss function doesn't factor over transitions meaningfully? (Since introducing an arc before a dependent has gathered all its dependents will lead to observed loss, even though this arc transition is augmented with loss all the time?)
			//E.g., we have gold and pred seqs with 10 transitions each. Pred contains 5 transitions that had its loss augmented by 3.5 in total. (this means pred has a score of 3.5 greater than w*fi(pred))
			//      However, these 5 lossaugmented transitions actually cause a greater loss in reality (when we apply the loss function offline), say 5.
			//
			//If tau = (p - g + l) / ( ||...||), where p denotes the actual score of the prediction, i.e., w*fi(pred), then we may be adding a too great loss here if we use what we get from the loss function.
			//
			//We can either let
			// l = oLoss, or
			// l = aLoss
			// where oLoss is "observed"  loss, i.e., what the loss function gives us (5 above), and
			//       aloss in "augmented" loss, i.e., what we have augmented the transitions with (3.5 above)
			double actPredScore=predFV.getScore(parameters.parameters);
			double oLoss=loss;
			double aLoss=predScore-actPredScore; //This is aLoss
			
//			predScore=actPredScore;
//			loss=aLoss;
//			loss=oLoss;
			
			//What's most funny, however, is that it seems like having a bigger numerator when computing tau seems to give better results.
			//I'm guessing the reason is that it reaches better numbers earlier, and that this works out better when doing averaging.
			//Now the real question is whether to just up the numerator of tau, or whether the loss in general should be increased (i.e., so that the loss augmentation makes it even more difficult to get things right during learning)
			//Here we first try by just playing with the loss a bit and do the following two things:
			//
			//1) Let predScore stay the same (with bonus for loss augmentation; this is what we had earlier, before correcting for the augmented loss in predScore)
			//2) Then we take the oLoss and double it (just to make up some number that makes tau bigger)
			
			
			//XXX There's something fishy going on here. It seems like it's useful to just make the numerator for the tau a bit bigger
			//    but I don't understand exactly what's going on here. I should look into this later. 
			
			//The following is my interpretation of how to properly do the lossaugmentation
//			{
//				predScore=actPredScore;
//				loss=oLoss; //This could be aLoss too, that's upp for discussion, given that we don't do exact search
//			}
			//This way was the inital bug (seems better)
//			{
//				predScore=predScore;
//				loss=oLoss;
//			}
			//Just making the numerator even bigger also seems to be better
//			{
//				predScore=predScore;
//				loss=2*oLoss;
//			}
		}
		
		IUpdStruct upd=parameters.getUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
		parameters.update(upd);
		return loss;
	}

	


	double getTransitionScore(State lastCorrect, ITransition correctTransition,Instance inst) {
//		long[] sink=new long[FeatureExtractor.SINK_MIN_SIZE];
		long[] sink=featureExtractor.sink;
		int f=featureExtractor.extractFeatures(lastCorrect, sink, inst);
		ArrayFunctions.shiftAllLeftByBits(sink, featureExtractor.bA, f);
		double transScore=decoder.getTransitionScore(correctTransition.getId(), sink, f);
		return transScore;
	}

	
	String getTrainInfo(int earlyBreak, int sentCount, long time, int correct, double accLoss, Bookkeeping bookkeeping) {
		double gs=(double)bookkeeping.goldSwaps/sentCount;
		double ps=(double)bookkeeping.predSwaps/sentCount;
		double gh=(double)bookkeeping.goldAssignedHeads/sentCount;
		double ph=(double)bookkeeping.predAssignedHeads/sentCount;
		double ogl=(double)bookkeeping.optGoldLen/sentCount;
		String s=String.format(" |EB: %-5d  c: %-5d t: %-2dms|l: %5.2f |gt: %6.3f  pt: %6.3f|u: %5d up: %5d |gs: %6.3f  ps: %6.3f|gh: %6.3f  ph: %6.3f|sf: %d|fda: %6.3f|ogl:%6.3f", earlyBreak,correct,time/sentCount,accLoss/sentCount,(double)bookkeeping.goldTrans/sentCount,(double)bookkeeping.predTrans/sentCount,bookkeeping.upds,bookkeeping.updPairs,gs,ps,gh,ph,l2i.signFlips(),(double)bookkeeping.forkDistSum/bookkeeping.updPairs,ogl);
		return s;
	}
	
	static class Bookkeeping {
		int goldTrans=0;
		int predTrans=0;
		int nullReturns=0;
		int upds=0;
		int updPairs=0;
		int goldAssignedHeads=0;
		int predAssignedHeads=0;
		int goldSwaps=0;
		int predSwaps=0;
		int forkDistSum=0;
		int optGoldLen=0;
		int sameTRSeqAsLast=0;
		int sameTRSeqAsCanon=0;
	}

	protected int doIteration(int beam, int itemCount) {
		long start=System.currentTimeMillis();
		DBO.printNoPrefix("S:");
		IFV<FV3> goldFV=new FV3();
		IFV<FV3> predFV=new FV3();
		int erase=0;
		int sentCount=0;
		int earlyBreak=0;
		int correct=0;
		double accLoss=0.d;

		Bookkeeping bookkeeping=new Bookkeeping();
		for(Instance inst:instances){
			oracle.initSentence(inst);
			itemCount++;

			State init=State.getInitialState(inst.head.length,options.labeled,inst.head); //heads are only needed for dynamic oracles, passing null reduces the memory footprint and copy cost

			List<Pair<State,State>> updates;
			switch(ut){
			case BeamEarly:             
				Pair<State,State> p=findEarlyUpdatePair(beam,inst,init);
				bookkeepLastPredTRSeq(p.getLeft(),inst,bookkeeping);
				updates=Collections.singletonList(p); 
				break;
			case BeamMaxVioGoldBeam:    
				updates=Collections.singletonList(findMaxVioUpdatePairGoldBeam(beam,goldBeam,bookkeeping,inst,init,goldFV,predFV)); 
				break;
			case BeamMaxVioGoldBeamOpt: 
				State[] gold=getGoldStateSeqBeamSearch(goldBeam, bookkeeping, inst, init);
				bookkeepLastPredTRSeq(gold[gold.length-1],inst,bookkeeping);
				updates=Collections.singletonList(findMaxVioUpdatePairGivenGold(beam,bookkeeping,inst,init,goldFV,predFV,gold)); 
				break;
			case BeamEarlyGoldBeam:     
			case BeamDelayLaSOGoldBeam:
			default: throw new Error("not implemented");
			}

			if(updates.size()==1){
				Pair<State,State> goldAndPred=updates.get(0);
				State gold=goldAndPred.getLeft();
				State pred=goldAndPred.getRight();
				if(DO_MEASURE_FORK_POINTS){
					bookkeeping.forkDistSum+=distToLastForkPoint(gold,pred);
				}
				bookkeeping.goldTrans+=gold.totalTransitions;
				bookkeeping.predTrans+=pred.totalTransitions;
				bookkeeping.goldAssignedHeads+=gold.getTotalAssignedHeads();
				bookkeeping.predAssignedHeads+=pred.getTotalAssignedHeads();
				bookkeeping.goldSwaps+=gold.countSwaps();
				bookkeeping.predSwaps+=pred.countSwaps();
				if(gold==pred){ //gold==pred is used when everything went right
					if(!pred.isTerminal())
						throw new Error("this shouldn't happen");
					++correct;
				} else {
					bookkeeping.updPairs++;
					bookkeeping.upds++;
					accLoss+=doUpdate(itemCount,goldFV,predFV,inst,pred,gold);
					if(!gold.isTerminal())
						earlyBreak++; 					//With swap it's less clear when an early break happened. Either when correct is terminal, or prediction is terminal (or both)
				}
			} else {
				int lgt=0;
				int lpt=0;
				int gt=0;
				int pt=0;

				int gs=0;
				int ps=0;
				int lgs=0;
				int lps=0;

				int gh=0;
				int ph=0;
				int lgh=0;
				int lph=0;

				goldFV.clear();
				predFV.clear();
				double pNorm=0.d;
				double gNorm=0.d;
				double loss=0.d;
				boolean finalCorrect=false;
				for(Pair<State,State> p:updates){
					State gold=p.getLeft();
					State pred=p.getRight();
					if(DO_MEASURE_FORK_POINTS){
						bookkeeping.forkDistSum+=distToLastForkPoint(gold,pred);
					}
					{ //track transition count
						gt+=gold.totalTransitions-lgt;
						pt+=pred.totalTransitions-lpt;
						lgt=gold.totalTransitions;
						lpt=pred.totalTransitions;
					}
					{ //track swap count
						int a=gold.countSwaps();
						int b=pred.countSwaps();
						gs+=a-lgs;
						ps+=b-lps;
						lgs=a;
						lps=b;
					}
					{ //track head count
						int a=gold.getTotalAssignedHeads();
						int b=pred.getTotalAssignedHeads();
						gh+=a-lgh;
						ph+=b-lph;
						lgh=a;
						lph=b;
					}
					if(gold==pred){
						finalCorrect=true;
						continue; //If we got the last bit right, then don't make updates from this
					}
					recoverFeaturesAndFillVector(goldFV, gold, inst, false);
					recoverFeaturesAndFillVector(predFV, pred, inst, false);
					loss+=lossFunction.getLoss(gold, pred, inst.form.length);
					gNorm+=gold.totalTransitions;
					pNorm+=pred.totalTransitions;
				}
				double predScore=predFV.getScore(parameters.parameters);
				double goldScore=goldFV.getScore(parameters.parameters);
				State lastCorr;
				State lastPred;
				if(finalCorrect){
					Pair<State, State> p=updates.get(updates.size()-2);
					lastCorr=p.getLeft();
					lastPred=p.getRight();
					bookkeeping.updPairs+=updates.size()-1;
				} else {
					Pair<State, State> p=updates.get(updates.size()-1);
					lastCorr=p.getLeft();
					lastPred=p.getRight();
					bookkeeping.updPairs+=updates.size();
				}


				//Try not scaling
				predScore/=pNorm;
				goldScore/=gNorm;
				//						predScore/=lastPred.totalTransitions;
				//						goldScore/=lastCorr.totalTransitions;
				predScore*=lastCorr.totalTransitions;
				goldScore*=lastCorr.totalTransitions;
				//					PerceptronUpdStruct upd=new PerceptronUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
				IUpdStruct upd=parameters.getUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
				parameters.update(upd);
				bookkeeping.goldTrans+=gt;
				bookkeeping.predTrans+=pt;
				bookkeeping.upds++;
				bookkeeping.goldSwaps+=gs;
				bookkeeping.predSwaps+=ps;
				bookkeeping.goldAssignedHeads+=gh;
				bookkeeping.predAssignedHeads+=ph;
				if(!lastCorr.isTerminal())
					earlyBreak++;
			}
			//				}
			sentCount++;
			if((sentCount-1)%111==0 || Options.DEBUG){
				erase=DBO.eraseAndPrint(erase, Integer.toString(sentCount));
				long time=System.currentTimeMillis()-start;
				erase+=DBO.printNoPrefix(getTrainInfo(earlyBreak,sentCount,time,correct,accLoss,bookkeeping));
				if(Options.DEBUG){
					DBO.println();
					erase=0;
				}
			}

		}
		//final printout
		erase=DBO.eraseAndPrint(erase, Integer.toString(sentCount));
		long time=System.currentTimeMillis()-start;
		DBO.printNoPrefix(getTrainInfo(earlyBreak,sentCount,time,correct,accLoss,bookkeeping));
		DBO.println();
		DBO.printlnNoPrefix(getTRSeqStats(bookkeeping,sentCount));
		//return item count
		return itemCount;
	}


	private State[] getGoldStateSeqBeamSearch(int goldBeam,Bookkeeping bookkeeping,Instance inst,State init){
//		State terminalGold=decoder.decodeGoldBeam(inst, goldBeam, oracle);
		State terminalGold=decoder.decodeGold(inst, goldBeam, oracle);
		State cur=terminalGold;
		bookkeeping.optGoldLen+=cur.totalTransitions;
		State[] seq=new State[cur.totalTransitions];
		for(int i=seq.length-1;i>=0;--i){
			seq[i]=cur;
			cur=cur.previous;
		}
		//Do a sanity heck here and see that we have the correct tree in terminalGold
		boolean correct=checkCorrectTree(terminalGold,inst);
		if(!correct)
			throw new Error("Gold beamed tree not correct?!");
		return seq;
	}
	
	private Pair<State, State> findMaxVioUpdatePairGivenGold(int beam,Bookkeeping bookkeeping, Instance inst, State init, IFV<FV3> goldFV,IFV<FV3> predFV, State[] goldSeq) {
		//Now do usual beam search wrt to this gold item
		
		Violation violation=new Violation();
		boolean returnOnEarly=Options.violationPolicy==ViolationPolicy.Early && Options.STRICT;
		boolean trackEarly=Options.violationPolicy==ViolationPolicy.Early || Options.violationPolicy==ViolationPolicy.HybridEarly;
		
		State terminalGold=goldSeq[goldSeq.length-1];
		
		Set<IntPair> goldSwaps=null;
		if(Options.lossFunctionType==LossFunction.Type.SwapHammingLabelWeighted)
			goldSwaps=terminalGold.getSwappedPairsInTRHistory();
		
		Pair<State,State> bestUpd=null;
		List<State> agenda=new ArrayList<State>(beam);
		agenda.add(init);
		
		MinHeap<TerminalState> terminalPreds=new MinHeap<TerminalState>(beam);
		for(int i=0;agenda.size()>0 && i<goldSeq.length;++i){
			State correctGold=goldSeq[i];
			List<State> newAgenda=decoder.advanceStates(agenda, beam, inst,Options.AUG_LOSS,goldSwaps);
			assignCorrectness(newAgenda,inst);
			

			State bestPred=newAgenda.get(0);

			if(!correctGold.equalTRSeq(bestPred)){ //Then gold and best pred are identical, to the very tr-seq
				double vio=getViolation(correctGold,bestPred);
				
				if(trackEarly && !violation.hasEarly()){
					boolean goldInBeam=containsEqualTRSeq(correctGold,newAgenda);
					if(!goldInBeam){
						violation.setEarlyUpdate(correctGold, bestPred, vio);
						if(returnOnEarly)
							return violation.early;
					}						
				}
				violation.offerVio(correctGold, bestPred, vio);
			}

			//Reseed agenda
			Decoder.filterExpansionsAndReseedAgenda(newAgenda, agenda, terminalPreds);
		}
		
		
		//Now continue the beam until it's done
		//Here we will have to compare to the terminal gold

		while(agenda.size()>0){
			List<State> newAgenda=decoder.advanceStates(agenda, beam, inst,Options.AUG_LOSS,goldSwaps);
//			List<State> newAgenda=decoder.advanceStates(agenda, beam, inst);
			assignCorrectness(newAgenda,inst);
			
			//Check violation on best
			State bestPred=newAgenda.get(0);
			double vio=getViolation(terminalGold,bestPred);
			
			if(bestPred.score()>terminalGold.score()){
				//Here we should get early update at the latest (if we enter this while loop at all)
				if(trackEarly && !violation.hasEarly()){
					//If early is not set yet, then we have to set it as soon as possible (i.e., first round of this loop)
					violation.setEarlyUpdate(terminalGold, bestPred, vio);
					if(returnOnEarly)
						return violation.early;
				}
				violation.offerVio(terminalGold, bestPred, vio);
			}

			//Reseed agenda
			Decoder.filterExpansionsAndReseedAgenda(newAgenda, agenda, terminalPreds);
		}

		State bestPred=MinHeap.extractMax(terminalPreds).state;
		
		//Sanity check on the oracle
		boolean correctTree=checkCorrectTree(bestPred,inst);
		if(bestPred.correct && !correctTree)
			throw new Error("here error -- tr sequence marked as correct but tree isn't");
		
		if(Options.STRICT){
			if(terminalGold.equalTRSeq(bestPred))
				return new Pair<State,State>(bestPred,bestPred);
		} else {
			if(correctTree)
				return new Pair<State,State>(bestPred,bestPred);
		}
		
		
		//XXX this was a bug in the IWPT paper on stagging/stacking (for the oracle paper we used max vio so it was not a problem)
		//To recreate the results from Falenska et al. 2015, comment out the if-block below
		//There is one corner case with early update:
		//-if the initial exploration of the beam (for loop above) terminates because agenda gets empty and we never saw all gold items
		//-but the gold item stayed in the beam all along
		//-then we don't have a proper early update, and the pair below (terminalGold vs bestTerminalPred) might not even be a violation
		//Here we add a work around:
		if(trackEarly && !violation.hasEarly() && bestPred.totalTransitions<terminalGold.totalTransitions){
			State g=goldSeq[bestPred.totalTransitions-1];
			double vio=getViolation(g,bestPred);
			violation.setEarlyUpdate(g, bestPred, vio);
		}
		
		//Now check vio on bestPred vs terminalGold too
		double vio=getViolation(terminalGold,bestPred);
		if(trackEarly && !violation.hasEarly()) //If we still don't have a first violation, now is the "earliest"
			violation.setEarlyUpdate(terminalGold, bestPred, vio);
		violation.offerVio(terminalGold, bestPred, vio); //Also offer the very final update -- terminal gold vs terminal pred
		
		
		boolean latestIsVio=violation.latestVio>0.d;
		switch(Options.violationPolicy){
		case Early:
			bestUpd=violation.early;
			break;
		case HybridEarly:
			bestUpd=latestIsVio?violation.latest:violation.early;
			break;
		case HybridMax:
			bestUpd=latestIsVio?violation.latest:violation.max;
			break;
		case Latest:
			bestUpd=violation.latest;
			break;
		case Max:
			bestUpd=violation.max;
			break;
		default:
			throw new Error("not implemented");
		}

		if(bestUpd!=null){
			if(bestUpd.getLeft().score()>bestUpd.getRight().score() && !Options.AUG_LOSS)
				DBO.println("\nBroken update pair: p:"+bestUpd.getRight().score()+", g: "+bestUpd.getLeft().score()+"\n");
			return bestUpd;
		} else {
			//and this is obviously correct, so return here.
			if(!bestPred.correct)
				throw new Error("best pred identical to best terminal gold but not marked as correct -- correctness marking issue?!");
			//TODO does this ever happen?
			System.err.println("\nRETURNING PSEUDO NULL -- look into this");
			return new Pair<State,State>(bestPred,bestPred);
		}

	}
	
	private double getViolation(State gold,State pred){
		return pred.score()-gold.score();
	}

	private boolean containsEqualTRSeq(State correctGold, List<State> newAgenda) {
		for(State s:newAgenda)
			if(correctGold.equalTRSeq(s))
				return true;
		return false;
	}


	private boolean checkCorrectTree(State bestPred, Instance inst) {
		return Arrays.equals(bestPred.treeSig,inst.head) && (bestPred.lblSig==null || Arrays.equals(bestPred.lblSig,inst.lbl));
	}

	private List<Pair<State, State>> beamDelayLaSO(int beam, Instance inst, State init) {
		List<Pair<State, State>> r=new ArrayList<Pair<State, State>>();
		State seed=init;
		do {
			Pair<State, State> p=findEarlyUpdatePair(beam, inst, seed);
			r.add(p);
			State corr=p.getLeft();
			if(corr.isTerminal())
				break;
			seed=corr;

		} while(true);
		return r;
	}
	
	
	
	private Pair<State,State> findMaxVioUpdatePairGoldBeam(int beam,int gbeam,Bookkeeping bookkeeping,Instance inst,State init, IFV<FV3> goldFV, IFV<FV3> predFV){
		//XXX what we need here is a fallback update -- something along the lines of "when the last gold item falls off the beam, find an update pair and save that in case maxvio doesn't fire anywhere"
		//See also the counter-example (on paper, in Anders' office) why the ideas outlined by Huang in the MT paper (Yu, Huang, Mi, Zhao, EMNLP 2013) with multiple golds are off and hence this is not a good idea
		
		int canLength=inst.transSequence.length;
		if(Options.DEBUG)
			System.out.println("\ncan length: "+canLength);
		List<State> goldAgenda=new ArrayList<State>();
		List<State> predAgenda=new ArrayList<State>();
		goldAgenda.add(init);
		predAgenda.add(init);
		
		double bestVio=-0.00001d;
		Pair<State,State> bestUpd=null;
		
		MinHeap<TerminalState> terminalGolds=new MinHeap<TerminalState>(gbeam);
		MinHeap<TerminalState> terminalPreds=new MinHeap<TerminalState>(beam);
		int gtb=0;
		int ptb=0;
		
		State bestTerminalGold=null;
		
		int i=0;
		
		while(predAgenda.size()>0 && goldAgenda.size()>0){
			++i;
			
			if(Options.DEBUG && i==canLength)
				System.out.println("i == canLength");
			List<State> newPredAgenda=decoder.advanceStates(predAgenda, beam, inst);
			List<State> newGoldAgenda=decoder.advanceGoldStates(goldAgenda, gbeam, inst, oracle);
			
			assignCorrectness(newPredAgenda,inst);
			assignCorrectness(newGoldAgenda,inst); //this should be redundant
			
			State bestPred=newPredAgenda.get(0);
			State bestGold=newGoldAgenda.get(0);
			if(bestTerminalGold!=null && bestTerminalGold.score()>bestGold.score())
				bestGold=bestTerminalGold;
			
			if(bestPred.correct){ //best prediction was correct
				//do nothing (?)
			} else { //best prediction was incorrect
				recoverFeaturesAndFillVector(goldFV,bestGold,inst,true);
				recoverFeaturesAndFillVector(predFV,bestPred,inst,true);
				double vio=parameters.reg.getViolation(parameters.parameters, goldFV.getDistVector(predFV));
				if(vio>bestVio){
					if(bestGold.score()>bestPred.score()){
						//this means the best item is outside the pruned search space, not much we can do...
						System.out.println("here");
					} else {
						bestUpd=new Pair<State,State>(bestGold,bestPred);
						bestVio=vio;
					}
				}
			}

			//Extract terminal and non-terminal states and reseed agenda
			for(State s:newGoldAgenda){
				if(s.isTerminal() && terminalGolds.offer(s.score())){
					if(bestTerminalGold==null || bestTerminalGold.score()<s.score())
						bestTerminalGold=s;
					terminalGolds.insert(new TerminalState(++gtb,s));
				}
			}
			for(State s:newPredAgenda){
				if(s.isTerminal() && terminalPreds.offer(s.score())){
					terminalPreds.insert(new TerminalState(++ptb,s));
				}
			}
			
			predAgenda.clear();
			goldAgenda.clear();
			
			for(State s:newGoldAgenda){
				if(!s.isTerminal() && (terminalGolds.hasRoomLeft() || terminalGolds.peek().score()<s.score())){
					goldAgenda.add(s);
				}
			}
			
			for(State s:newPredAgenda){
				if(!s.isTerminal() && (terminalPreds.hasRoomLeft() || terminalPreds.peek().score()<s.score())){
					predAgenda.add(s);
				}
			}
			
		}
		if(Options.DEBUG)
			System.out.println("Last i: "+i);
		
		//Now we fell out of this loop either because
		//1) gold agenda empty
		//2) pred agenda empty
		//3) both agendas empty
		//Handle these cases now
		
		if(predAgenda.isEmpty() && goldAgenda.isEmpty()){
			//Here we are safe -- do nothing
		} else if(predAgenda.isEmpty()) {
			//Here we are also safe -- only more gold items left, but we don't care
		} else if(goldAgenda.isEmpty()){
			//This is the case when we can go on with predictions
			//We keep doing this until the prediction agenda dies naturally
			//Since all gold items are shorter, we will compare with the best gold item altogether (bestTerminalGold)
			
			recoverFeaturesAndFillVector(goldFV,bestTerminalGold,inst,true);
			while(predAgenda.size()>0){
				List<State> newAgenda=decoder.advanceStates(predAgenda, beam, inst);
				State bestPred=newAgenda.get(0);
				double vio=parameters.reg.getViolation(parameters.parameters, goldFV.getDistVector(predFV));
				if(vio>bestVio){
					if(bestTerminalGold.score()>bestPred.score()){
						//this means the best item is outside the pruned search space, not much we can do...
						System.out.println("here: "+(bestTerminalGold.score()-bestPred.score()));
					} else {
						bestUpd=new Pair<State,State>(bestTerminalGold,bestPred);
						bestVio=vio;
					}
				}
				//reseed agenda
				predAgenda.clear();
				for(State s:newAgenda){
					if(s.isTerminal()){
						if(terminalPreds.offer(s.score()))
							terminalPreds.insert(new TerminalState(++ptb,s));
					} else {
						predAgenda.add(s);
					}
				}
			}
		}
		
		State bestTerminalPred=MinHeap.extractMax(terminalPreds).state;
		
		if(Options.LAZY_UPDATE && bestTerminalGold!=null && bestTerminalPred.equalTRSeq(bestTerminalGold)) //here we made the correct prediction
			return new Pair<State,State>(bestTerminalPred,bestTerminalPred);
		
		if(Options.LAZY_UPDATE && bestTerminalPred.correct) //Lazyness, if the best terminal was still correct, then do nothing
			return new Pair<State,State>(bestTerminalPred,bestTerminalPred);
		
		if(bestUpd==null){
			if(bestTerminalPred.correct)
				return new Pair<State,State>(bestTerminalPred,bestTerminalPred);
			System.err.println("\nNO BEST UPDATE FOUND and INCORRECT PREDICTION");
			return new Pair<State,State>(bestTerminalGold,bestTerminalGold);
		} else {
			return bestUpd;
		}
		
		
//		
//
//		State bestTerminalGold=null;
//		State bestTerminalPred=null;
//		if(terminalPreds.size()>0)
//			bestTerminalPred=MinHeap.extractMax(terminalPreds).state;
//		if(terminalGolds.size()>0)
//			bestTerminalGold=MinHeap.extractMax(terminalGolds).state;
//		
//		if(bestTerminalPred!=null && bestTerminalPred.correct)
//			return new Pair<State,State>(bestTerminalPred,bestTerminalPred);
//		else if(bestUpd==null){
//			//This is a tricky boundary case we need to deal with.
//			State g=bestTerminalGold;
//			State p=bestTerminalPred;
//			
//			String gs="null";
//			String ps="null";
//			if(g!=null)
//				gs=String.format("%6f / %6f", g.totalScore,g.score());
//			if(p!=null)
//				ps=String.format("%6f / %6f", p.totalScore,p.score());
//			
//			System.err.printf("\nBoundary case: ga: %d, pa: %d    ||    g: %s | p: %s\n",goldAgenda.size(),predAgenda.size(),gs,ps);
//			if(g!=null && p!=null)
//				return new Pair<State,State>(g,p);
//			else if(g!=null)
//				return new Pair<State,State>(g,g); //This means no update... not ideal
//			else if(p!=null)
//				return new Pair<State,State>(p,p); //This means no update... not ideal
//			else
//				throw new Error("nothing terminal!?!?");
//		}else
//			return bestUpd;
	}

	/*
	 * this is superseeded by the findMaxVioUpdatePairGivenGold
	 */
	@Deprecated
	private Pair<State, State> findMaxVioUpdatePairStatic(int beam, Bookkeeping bookkeeping, Instance inst, State init, IFV<FV3> goldFV, IFV<FV3> predFV) {
		if(oracle.isDynamic())
			throw new Error("dynamic oracle with BeamMaxVio is not implemented"); //Then we need to keep going and keep the one-best dynamic sequence around, whereas this method uses the canonical sequence only
		State correct=init.deepCopy();
		State bestPred=null;
		State bestTerminalPred=null;
		int c=0;
		List<State> agenda=Collections.singletonList(init);
		double bestVio=-9999.d;
		Pair<State,State> bestUpd=null;
		while(agenda.size()>0){
			List<State> newAgenda=decoder.advanceStates(agenda, beam, inst);
			assignCorrectness(newAgenda,inst);
			if(c<inst.transSequence.length)
				correct=decoder.applyTransition(inst.transSequence[c++], correct, inst);
			bestPred=newAgenda.get(0);
			//Reseed agenda
			agenda=new ArrayList<State>();
			for(State s:newAgenda)
				if(!s.isTerminal())
					agenda.add(s);
			if(bestPred.correct){ //Then we're safe, no need to update
				if(bestPred.isTerminal() && (bestTerminalPred==null || bestTerminalPred.score()<bestPred.score())){
					bestTerminalPred=bestPred;
				}
				continue;
			}
			
			//If we fall down here bestPred is incorrect. But we might have a terminal best prediction that has a higher score, in which case we shouldn't compare with this guy
			if(bestTerminalPred!=null && bestTerminalPred.score()>bestPred.score())
				continue;

			recoverFeaturesAndFillVector(goldFV,correct,inst,true);
			recoverFeaturesAndFillVector(predFV,bestPred,inst,true);

			double vio=parameters.reg.getViolation(parameters.parameters, goldFV.getDistVector(predFV));
			if(vio>bestVio){
				bestUpd=new Pair<State,State>(correct,bestPred);
				bestVio=vio;
			}
		}
		if(bestUpd==null){
			bookkeeping.goldTrans+=correct.totalTransitions;
			bookkeeping.predTrans+=bestPred.totalTransitions;
		}
		return bestUpd;
	}
	
	private Pair<State, State> findEarlyUpdatePairNEW(int beam, Instance inst, State init) {
		List<State> agenda=new ArrayList<State>(beam);
		agenda.add(init);
		State lastCorrect=init;
		State bestTerminalCorrect=null;
		State bestTerminalPred=null;
		MinHeap<TerminalState> terminals=new MinHeap<TerminalState>(beam);
		while(agenda.size()>0){
			List<State> expansions=decoder.advanceStates(agenda, beam, inst);
			assignCorrectness(expansions,inst);
			Decoder.filterExpansionsAndReseedAgenda(expansions, agenda, terminals);
		}
		
		//XXX TODO
		
		return null;
	}
	
	private Pair<State, State> findEarlyUpdatePair(int beam, Instance inst, State init) {
		List<State> agenda=Collections.singletonList(init);
		State lastCorrect=init;
		State terminalCorrect=null;
		State bestTerminalPred=null;
		while(agenda.size()>0){
			List<State> newAgenda=decoder.advanceStates(agenda, beam, inst);
			assignCorrectness(newAgenda,inst);
			State bestCorrect=extractBestCorrect(newAgenda);
			State bestPred=newAgenda.get(0);
			if(bestCorrect!=null && !bestCorrect.correct) //sanity check
				System.err.println("Here");

			//Now we have a couple of possible outcomes if there is no correct item in the beam
			//0: There is a correct item in the beam, then we just go on
			//1: The correct item fell off the beam, but we weren't done yet
			//2: We already found the correct *terminal* previously (and because of swaps we've gone past it), then we have it in terminalCorrect
			//3: The best pred in newAgenda has fewer transitions but is already terminal. In the worst case we will have the whole newAgenda full of terminal items that have fewer transitions than the correct. Then what do we update against?
			if(bestCorrect!=null){ //Handle 0
				//Store terminal correct
				//if we're using a dynamic oracle, then we might've found a better one, then we update
				if(bestCorrect.isTerminal()){
					if(terminalCorrect==null || terminalCorrect.score()<bestCorrect.score())
						terminalCorrect=bestCorrect;
				}
				lastCorrect=bestCorrect;
				
			} else if(terminalCorrect==null){ //Handle 1
				State correctState=recoverCorrect(inst,lastCorrect,bestPred);
				if(bestTerminalPred!=null && bestTerminalPred.score()>bestPred.score()){
					bestPred=bestTerminalPred;
				}
					
				return new Pair<State,State>(correctState,bestPred);
			} else { //Handle 2
//				agenda=reseedAgenda(newAgenda,terminalPreds);
//				agenda=newAgenda;
//				agenda=Decoder.extractNonTerminal(newAgenda);
			}
			
			//Track the very best terminal prediction we've made so far
			State bt=Decoder.extractBestTerminal(newAgenda);
			if(bestTerminalPred==null){
				bestTerminalPred=bt;
			} else if(bt!=null){
				if(bt.score()>bestTerminalPred.score())
					bestTerminalPred=bt;
			}
			agenda=Decoder.extractNonTerminal(newAgenda);			
		}
		//Final check at the end of the instance
		State bestPred=bestTerminalPred;
//		if(bestTerminalPred!=null && bestTerminalPred.getAvgScore()>bestPred.getAvgScore())
//			bestPred=bestTerminalPred;
		State correctState;
		if(terminalCorrect==null){
			correctState = recoverCorrect(inst,lastCorrect,bestPred);
		} else {
			correctState = terminalCorrect;
		}
		if(!bestPred.correct){
			return new Pair<State,State>(correctState,bestPred);
		} else { //Need to track this when we don't return and make an update
			return new Pair<State,State>(bestPred,bestPred);
		}
	}

	private void assignCorrectness(List<State> newAgenda,Instance inst) {
		for(State s:newAgenda)
			if(s.correct && isCorrectAction(s.lastAction,s.previous,inst))
				s.correct=true;
			else
				s.correct=false;
	}

	private State recoverCorrect(Instance inst, State lastCorrect, State pred) {
		if(pred.totalTransitions<=lastCorrect.totalTransitions){
			//Then we just return last correct
			return lastCorrect;
		} else if(lastCorrect.totalTransitions!=pred.totalTransitions-1){ //This would be the usual case, when the correct item fell off the beam. But with swap, the best pred could have come earlier (and then above happens)
			System.out.println("Last correct: "+lastCorrect.totalTransitions);
			System.out.println("Pred:         "+pred.totalTransitions);
			throw new Error("I think this can happen with swap, but not without");
		}
		ITransition best=null;
		double bestScore=-9999.d;
		List<ITransition> ok=getCorrectTransitions(lastCorrect,inst);
		if(ok.isEmpty()){
			System.err.println("No correct??");
			ok=getCorrectTransitions(lastCorrect,inst);
		}
		for(ITransition tr:ok){
			double transScore=getTransitionScore(lastCorrect,tr,inst);
			if(transScore>bestScore){
				bestScore=transScore;
				best=tr;
			}
		}
		if(best==null){
			System.err.println("NO BEST ???");
		}
		
		State correctState=best.apply(lastCorrect,bestScore+lastCorrect.totalScore);
		return correctState;
	}

	private State extractBestCorrect(List<State> newAgenda) {
		State best=null;
		for(State s:newAgenda)
			if(s.correct && (best==null || best.score()<s.score()))
				best=s;
		return best;
	}
	
	abstract boolean isCorrectAction(ITransition lastAction, State previous, Instance inst);
	abstract List<ITransition> getCorrectTransitions(State lastCorrect,	Instance inst);

	
	static int distToLastForkPoint(State gold,State pred){
		ITransition[] gSeq=State.getTransitionSequence(gold);
		ITransition[] pSeq=State.getTransitionSequence(pred);
		
		int min=Math.min(gSeq.length, pSeq.length);
		int max=Math.max(gSeq.length, pSeq.length);
		int com=0;
		
		while(com<min && gSeq[com]==pSeq[com])
			++com;
		
		
		return max-com;
	}
	
	class Violation {
		Pair<State,State> early;
		Pair<State,State> max;
		Pair<State,State> latest;
		
		double earlyVio=0.d;
		double maxVio=0.d;
		double latestVio=0.d;
		
		public void setEarlyUpdate(State gold,State pred,double vio){
			if(gold.score()>pred.score()){
				System.err.println("\n\nSetting bad early update pair:\n\n");
				System.err.println("p: "+pred.score()+", g:"+gold.score());
				System.err.println();
//				throw new Error("!");
			}
			if(early!=null)
				throw new Error("can't set early twice");
			Pair<State,State> p=new Pair<State,State>(gold,pred);
			this.early=p;
			this.earlyVio=vio;
		}
		
		public void offerVio(State gold,State pred,double vio){
			Pair<State,State> p=new Pair<State,State>(gold,pred);
			if(latest==null || vio>0.d){
				latest=p;
				latestVio=vio;				
			}
			if(max==null || vio>maxVio){
				max=p;
				maxVio=vio;
			}
		}
		
		public boolean hasEarly(){
			return early!=null;
		}
	}
}
