package ims.trans;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.ReaderWriterFactory;
import ims.oracle.IOracle;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.trans.AbstractBeamTrainer.Bookkeeping;
import ims.trans.Options.UpdateType;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.util.Eval;
import ims.trans.util.TRSeqPathAnalysis;
import ims.util.Accuracy;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.IOUtils;
import ims.util.Pair;
import ims.util.Util;

public abstract class AbstractTrainer implements ITrainer {

	protected abstract int doIteration(int beam, int itemCount);
	
	protected final Options options;
	protected final SymbolTable symTab;
	protected final IOracle oracle;
	protected final ITransitionFilter transitionFilter;
	protected final List<Instance> instances;
	protected final FeatureExtractor featureExtractor;
	protected final Long2IntInterface l2i;
	protected final ParametersFloat parameters;
	protected final Decoder decoder;
	protected final UpdateType ut;
	protected final InstanceMapping instanceMapping;
	protected final boolean mergeEquivalentStates;
	protected final LossFunction lossFunction;
	
	public AbstractTrainer(Options options, SymbolTable symTab, IOracle oracle,ITransitionFilter transitionFilter, List<Instance> instances,FeatureExtractor featureExtractor, Long2IntInterface l2i,ParametersFloat parameters,Decoder decoder,LossFunction lossFunction,InstanceMapping instanceMapping) {
		this.options = options;
		this.symTab = symTab;
		this.oracle = oracle;
		this.transitionFilter = transitionFilter;
		this.instances = instances;
		this.featureExtractor = featureExtractor;
		this.l2i = l2i;
		this.parameters = parameters;
		this.decoder = decoder;
		this.mergeEquivalentStates=options.mergeEquivalentStates;
		this.ut=options.ut;
		this.lossFunction=lossFunction;
		this.instanceMapping=instanceMapping;
	}
	
	protected void doInterIterTest(int itemCount) throws IOException, UnsupportedEncodingException, FileNotFoundException {
		if(options.testEveryIter){
			l2i.freeze();
			float[] params;
			if(Options.useAveraging)
				params=parameters.averagedCopy(itemCount);
			else
				params=parameters.parameters;
			Decoder d=new Decoder(params, l2i, symTab, transitionFilter, featureExtractor,options.mergeEquivalentStates,options.labeled,instanceMapping);
			DBO.printNoPrefix("Testing on ");
			Test.applyDecoder(d,ReaderWriterFactory.getReader(options.test,options.format.sentenceClass,options.reversed),ReaderWriterFactory.getWriter(options.out,options.reversed),options.beam);
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(options.out, options.gold, options.format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
			l2i.unFreeze();
		}
	}

	public Model train() throws IOException {
		int itemCount=0;
		for(int iter=0;iter<options.iterations;++iter){
			DBO.println("Iteration "+iter+" \t");
			boolean threadSafe=!l2i.frozen();
			itemCount=doIteration(options.beam,itemCount);
			printNonZero(parameters);
			doInterIterTest(itemCount);
			oracle.nextIter();
			if(Options.shuffleStrategy.shuffleBetweenIterations)
				ArrayShuffler.SINGLETON.shuffle(instances);
			if(Options.outTRFilePF!=null)
				dumpTRSeqs(iter);
		}

		if(Options.useAveraging)
			parameters.average(itemCount);
		
		return new Model(featureExtractor, symTab, parameters.parameters, l2i, transitionFilter,options.beam,options.mergeEquivalentStates,options.labeled,instanceMapping,options.transitionSystem,options.reversed);
	}

	private void dumpTRSeqs(int iter) throws IOException {
		File outFile=TRSeqPathAnalysis.getFileName(Options.outTRFilePF, iter);
		BufferedWriter writer=IOUtils.getBufferedWriter(outFile);
		for(Instance inst:instances){
			List<String> l=new ArrayList<String>();
			for(int i:inst.lastTRSeq)
				l.add(symTab.transition2label[i].getFullTransitionLabel());
			writer.write(inst.sentId+":");
			writer.write(Util.joinStrings(l, ","));
			writer.newLine();
		}
		writer.close();
	}

	private void printNonZero(ParametersFloat parameters) {
		int nz=parameters.countNZ();
		int total=parameters.total.length;
		int l=Integer.toString(total).length();
		String format="Parameters: nz: %"+l+"d  / total: %"+l+"d";
		DBO.println(String.format(format,nz,total));
	}

	protected void bookkeepLastPredTRSeq(State gold, Instance inst, Bookkeeping bookkeeping) {
		int[] newSeq;
		if(gold.isTerminal()){
			newSeq=new int[gold.totalTransitions];
		} else {
			newSeq=new int[gold.totalTransitions+1];
			newSeq[gold.totalTransitions]=-1; //sentinel for marking incomplete sequence (for early update)
		}
		State c=gold;
		for(int i=gold.totalTransitions-1;i>=0;--i){
			newSeq[i]=c.lastAction.getId();
			c=c.previous;
		}
		
		//Check if we have the same (partial) seq as last time
		boolean sameSeqAsLast=true;
		if(inst.lastTRSeq==null){
			sameSeqAsLast=false;
		} else {
			for(int i=0,m=Math.min(newSeq.length, inst.lastTRSeq.length);i<m && inst.lastTRSeq[i]!=-1 && newSeq[i]!=-1;++i){
				if(inst.lastTRSeq[i]!=newSeq[i]){
					sameSeqAsLast=false;
					break;
				}
			}
		}
		
		if(sameSeqAsLast)
			bookkeeping.sameTRSeqAsLast++;
		
		//Check if we have the same (partial) seq as canon
		boolean sameAsCanon;
		if(newSeq.length>inst.transSequence.length){
			sameAsCanon=false;
		} else {
			sameAsCanon=true;
			for(int i=0;i<newSeq.length;++i){
				if(newSeq[i]==-1 && i==newSeq.length-1)
					break;
				if(newSeq[i]!=inst.transSequence[i]){
					sameAsCanon=false;
					break;
				}
			}
		}
		
		if(sameAsCanon){
			bookkeeping.sameTRSeqAsCanon++;
		}
		
		inst.lastTRSeq=newSeq;
	}

	protected String getTRSeqStats(Bookkeeping bookkeeping,int sentCount) {
		double goldIdenticalToLast=100.0*bookkeeping.sameTRSeqAsLast/sentCount;
		double goldIdenticalToCanon=100.0*bookkeeping.sameTRSeqAsCanon/sentCount;
		return String.format("GSeqLast: %6.2f  |  GSeqCanon: %6.2f", goldIdenticalToLast,goldIdenticalToCanon);
	}

}
