package ims.trans;

import java.util.Set;

import ims.trans.transitions.ITransition;
import ims.util.EditDistance;
import ims.util.IntPair;

public abstract class LossFunction {

	public static final double LOSS_SCALE_FACTOR=1.d;
	
	public static final double SWAP_FP_LOSS=1.d*LOSS_SCALE_FACTOR;
	public static final double SWAP_FN_LOSS=1.d*LOSS_SCALE_FACTOR;
	public static final double ARC_FP_LOSS=1.d*LOSS_SCALE_FACTOR;
	public static final double ARC_L_FP_LOSS=0.5d*LOSS_SCALE_FACTOR;
	
	public enum Type { Hamming, HammingLabelWeighted, Levenshtein, LevenshteinLabelWeighted, HammingLabelWeightedAndSwap, HammingLabelWeightedAndSwapBiDir, None, SwapHammingLabelWeighted }

	public abstract double getLoss(State corr, State pred, int length);

	public static LossFunction getLossFunction(Type lossFunctionType) {
		switch(lossFunctionType){
		case None:                        return new None();
		case Hamming:                     return new Hamming();
		case HammingLabelWeighted:        return new HammingLabelWeighted();
		case Levenshtein:                 return new Levenshtein();
		case LevenshteinLabelWeighted:    return new LevenshteinLabelWeighted();
		case HammingLabelWeightedAndSwap: return new HammingLabelWeightedAndSwap();
		case HammingLabelWeightedAndSwapBiDir: return new HammingLabelWeightedAndSwapBiDir();
		case SwapHammingLabelWeighted:    return new SwapHammingLabelWeighted();
		default: throw new Error("not implemented");
		}
	}
	
	static class None extends LossFunction {

		@Override
		public double getLoss(State corr, State pred, int length) {
			return 0;
		}
		
	}
	
	static class SwapHammingLabelWeighted extends LossFunction {
		private final HammingLabelWeighted hlw=new HammingLabelWeighted();
		
		@Override
		public double getLoss(State corr, State pred, int length) {
			double r=hlw.getLoss(corr, pred, length);
			//Now we need to compute false positive and false negative swaps
			Set<IntPair> predSwaps=pred.getSwappedPairsInTRHistory();
			Set<IntPair> goldSwaps=corr.getSwappedPairsInTRHistory();
			
			int ps=predSwaps.size();
			int gs=goldSwaps.size();
			
			predSwaps.retainAll(goldSwaps); //Now predSwaps is the intersection
			int tp=predSwaps.size();
			
			int fp=ps-tp; //False positives are swaps we did but we shouldn't have done
			int fn=gs-tp; //False negatives are swaps we should have done but didn't do
			
			r+=fp*SWAP_FP_LOSS;
			r+=fn*SWAP_FN_LOSS;
			
			return r;
		}

		
	}
	
	static class HammingLabelWeightedAndSwapBiDir extends LossFunction {
		private final HammingLabelWeighted hlw=new HammingLabelWeighted();

		@Override
		public double getLoss(State corr, State pred, int length) {
			double r=hlw.getLoss(corr, pred, length);
			double gSwaps=getSwapCount(corr);
			double pSwaps=getSwapCount(pred);			
			double sw=Math.abs(pSwaps-gSwaps);
			
			return r+sw;
		}
	}
	
	static class HammingLabelWeightedAndSwap extends LossFunction {

		private final HammingLabelWeighted hlw=new HammingLabelWeighted();
		
		@Override
		public double getLoss(State corr, State pred, int length) {
			double r=hlw.getLoss(corr, pred, length);
			double gSwaps=getSwapCount(corr);
			double pSwaps=getSwapCount(pred);			
			double sw=Math.max(0, pSwaps-gSwaps); //This is the number of swaps that Pred did *more* than Gold. (so 0 if we swapped less, but then there will be loss in the hlw loss, since we must've made more arcs)
			
			return r+sw;
		}

	}
	
	static class Hamming extends LossFunction {
		@Override
		public double getLoss(State corr, State pred, int length) {
			double l=0.d;
			for(int i=1;i<length;++i){
				int g=corr.getHead(i);
				int p=pred.getHead(i);
				if(g!=p)
					l++;
			}
			return l;
		}
	}

	static class HammingLabelWeighted extends LossFunction {
		@Override
		public double getLoss(State corr, State pred, int length) {
			double l=0.d;
			for(int i=1;i<length;++i){
				int g=corr.getHead(i);
				int p=pred.getHead(i);
				if(g!=p){
					l+=ARC_FP_LOSS;
				} else {
					int gl=corr.getLabel(i);
					int pl=pred.getLabel(i);
					if(gl!=pl)
						l+=ARC_L_FP_LOSS;
				}
			}
			return l;
		}		
	}
	
	static class Levenshtein extends LossFunction {
		@Override
		public double getLoss(State corr, State pred, int length) {
			return EditDistance.levenshteinDistance(corr.getTransitionSeqSignature(), pred.getTransitionSeqSignature());
		}
	}
	
	static class LevenshteinLabelWeighted extends LossFunction {
		@Override
		public double getLoss(State corr, State pred, int length) {
			ITransition[] g=State.getTransitionSequence(corr);
			ITransition[] p=State.getTransitionSequence(pred);
			double ulbl=EditDistance.levenshteinDistance(corr.getTransitionSeqSignature(), pred.getTransitionSeqSignature());
			double lbl=EditDistance.levenshteinDistance(g, p);
			if(lbl<ulbl)
				throw new Error("this shouldn't happen...");
			return ulbl+(lbl-ulbl)/2;
		}
	}
	
	static double getSwapCount(State state) {
		return state.countSwaps();
	}
	
}
