package ims.trans;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.zip.ZipOutputStream;

import ims.data.Instance;
import ims.data.ISentence;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.NonProjDiscardingReader;
import ims.io.ReaderWriterFactory;
import ims.io.ReversingReader;
import ims.io.SentenceReader;
import ims.oracle.AbstractDynamicOracle;
import ims.oracle.DeterministicFileOracle;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.perceptron.Long2Int;
import ims.perceptron.Long2IntExact;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.Regularizer;
import ims.trans.CVPrintTRSeq.CVStruct;
import ims.trans.Options.Oracle;
import ims.trans.Options.TransitionSystem;
import ims.trans.features.FeatureExtractorFactory;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFilterFactory;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.NonProjInfo;
import ims.util.ZipUtil;

public class Train {
	
	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		options.sanityCheckTestEvery();
		options.warnNoModel();
		
		if(options.testEveryIter && options.out==null){
			System.err.println("-testEvery but no -out given, stopping.");
			System.exit(1);
		}
		
		SentenceReader<?> reader = getTrainReader(options);
		
		IOracle oracle=OracleFactory.getOracle(options);
		
		oracle = sanityCheckProjectivity(options, reader, oracle);

		SymbolTable symTab=new SymbolTable(options.useStag,true);
		
		InstanceMapping instanceMapping=new InstanceMapping(options.useStag,options.useStack);

		List<Instance> instances = createInstances(options, reader, oracle, symTab,	instanceMapping);
		
		if(options.oracle==Oracle.File){
			((DeterministicFileOracle) oracle).initSeqsFromFile(options.trSeqFile, instances, symTab);
		}
		
		if(options.cvSharedSymTab){
			trainCVShared(options,oracle,symTab,instanceMapping,instances);
		} else {
			trainNormal(options, oracle, symTab, instanceMapping, instances);
		}
		
		options.done();
	}

	private static void trainCVShared(Options options, IOracle oracle,SymbolTable symTab, InstanceMapping instanceMapping,List<Instance> instances) throws IOException {
		if(options.model==null){
			System.err.println("Model set to null for CVShare training, aborting...");
			System.exit(1);
		}
		//Ok, start by partitioning the instances for each job
		int f=options.folds;
		List<List<Instance>> trainingData=new ArrayList<List<Instance>>();
		List<List<Instance>> folds=cvSplit(instances,f);

		for(int i=0;i<f;++i){
			List<Instance> t=new ArrayList<Instance>();
			for(int j=0;j<f;++j){
				if(i==j)
					continue;
				List<Instance> fold=folds.get(j);
				t.addAll(fold);
			}
			trainingData.add(t);
		}
		
		//Bookkeeping stuff
		int[] sOffsets=new int[f];
		for(int i=0,t=0;i<f;++i){
			t+=folds.get(i).size();
			sOffsets[i]=t;
		}
		int[] sLengths=new int[instances.size()];
		for(int i=0;i<instances.size();++i)
			sLengths[i]=instances.get(i).form.length;		
		
		//Ok, now create f AbstractTrain objects
		List<AbstractTrainer> trainJobs=new ArrayList<AbstractTrainer>();
		
		//Implement this first single threaded
		//Then add a switch to parallelize it using a quiet switch in the trainer to prevent the output from getting scrambled.
		//XXX none of the oracles are threadsafe, so we should probably create one oracle for each job during parallelization
		for(int i=0;i<f;++i){
			AbstractTrainer t=getTrainer(options, oracle, symTab, instanceMapping, trainingData.get(i));
			trainJobs.add(t);
		}
		
		List<String> modelInfo = getModelInfo(options, oracle, trainJobs.get(0));
		modelInfo.add("");
		modelInfo.add("CV Stuff:");
		modelInfo.add("Folds:            "+f);
		modelInfo.add("SentenceOffsets:  "+Arrays.toString(sOffsets));
		printModelInfo(modelInfo);
		
		List<Model> models=new ArrayList<Model>();
		
		for(AbstractTrainer t:trainJobs){
			Model m=t.train();
			models.add(m);
		}
		
		CVStruct cvS=new CVStruct(f, sLengths, sOffsets, models);
		DBO.println("Done training, saving model");
		ZipOutputStream zos=ZipUtil.openZOS(options.model);
		ZipUtil.writeObjectAsZipEntry(zos, CVStruct.ZIP_FILE_ENTRY, cvS);
		ZipUtil.writeStringTXT(zos, "info.txt", modelInfo);
		zos.close();
		DBO.println("Done.");
	}

	private static void trainNormal(Options options, IOracle oracle, SymbolTable symTab, InstanceMapping instanceMapping, List<Instance> instances) throws IOException {
		if(Options.shuffleStrategy.shuffleFirstIteration){
			DBO.println("Shuffling instances");
			new ArrayShuffler().shuffle(instances);
		}
		
		double avgTransSeqLen=getAvgTransSeqLen(instances);
		DBO.printlnNoPrefix("Average transition sequence length (canonical):   "+String.format("%8.2f",avgTransSeqLen));
		//Train object
		AbstractTrainer train=getTrainer(options, oracle, symTab, instanceMapping,instances);
		
		//Now start training
		List<String> modelInfo = getModelInfo(options, oracle, train);
		printModelInfo(modelInfo);
		
		Model m=train.train();
		
		m.l2i.freeze();
		
		if(options.model!=null){
			DBO.println("Done training, saving model");
			ZipOutputStream zos=ZipUtil.openZOS(options.model);
			ZipUtil.writeObjectAsZipEntry(zos, Model.ZIP_FILE_ENTRY, m);
			ZipUtil.writeStringTXT(zos, "info.txt", modelInfo);
			zos.close();
			DBO.println("Done.");
		} else {
			DBO.println("Done training, SKIPPING saving model");
		}
	}

	protected static void printModelInfo(List<String> modelInfo) {
		DBO.printlnNoPrefix("Model Info:");
		for(String info:modelInfo)
			DBO.printlnNoPrefix(info);
		DBO.println();
		DBO.println("Now begin training");
	}

	protected static List<String> getModelInfo(Options options, IOracle oracle,AbstractTrainer train) {
		List<String> r=new ArrayList<String>();
		r.add("Oracle:       "+options.oracle);
		r.add("System:       "+oracle.getClass().getCanonicalName());
		r.add("StatSys:      "+Options.statSystem);
		r.add("DynSys:       "+Options.dynSystem);
		r.add("Merge States: "+options.mergeEquivalentStates);
		r.add("Beam:         "+options.beam);
		r.add("Reg:          "+train.parameters.reg.toString());
		r.add("Iterations:   "+options.iterations);
		r.add("HSize:        "+train.decoder.l2i.size());
		r.add("LossFunction: "+train.lossFunction.getClass().getCanonicalName());
		r.add("FS:           "+train.decoder.featureExtractor.getClass().getCanonicalName());
		r.add("UseStag:      "+options.useStag);
		r.add("UseStack:     "+options.useStack);
		r.add("GreedyTrain:  "+options.trainGreedy);
		r.add("Reversed:     "+options.reversed);
		r.add("Shuffle:      "+Options.shuffleStrategy);
		r.add("Averaging:    "+Options.useAveraging);
		r.add("Upd:          "+options.ut);
		r.add("Vp:           "+Options.violationPolicy);
		return r;
	}

	protected static AbstractTrainer getTrainer(Options options,IOracle oracle, SymbolTable symTab,InstanceMapping instanceMapping, List<Instance> instances) {
		LossFunction lossFunction=LossFunction.getLossFunction(Options.lossFunctionType);
		Regularizer regularizer=Regularizer.getRegularizer(options.regularizerType, options.C,options.eta,options.lambda);
		Long2IntInterface l2i=options.useHashKernel?new Long2Int(options.hashSize):new Long2IntExact(options.hashSize);		
		ParametersFloat parameters=new ParametersFloat(l2i.size(),regularizer);
		ITransitionFilter transitionFilter=TransitionFilterFactory.getTransitionFilter(symTab, instances, options, oracle);
		FeatureExtractor featureExtractor=FeatureExtractorFactory.getExtractor(symTab,options);
		Decoder decoder=new Decoder(parameters.parameters, l2i, symTab, transitionFilter,featureExtractor,options.mergeEquivalentStates,options.labeled,instanceMapping);
		if(options.initBias)
			featureExtractor.initWeights(symTab, parameters, l2i);
		
		
		final AbstractTrainer train;
		if(options.trainGreedy){
			train=new GreedyTrainer(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping);
			options.beam=1;
		} else {
			if(options.oracle==Oracle.Dynamic)
				train=new DynamicOracleTrainer(options, symTab, (AbstractDynamicOracle) oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping,Options.goldBeam);
			else if(options.oracle==Oracle.File)
				train=new FileStaticOracleTrainer(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping,Options.goldBeam);
			else
				train=new StaticOracleTrainer(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping,Options.goldBeam);
		}
		return train;
	}

	static List<Instance> createInstances(Options options,SentenceReader<?> reader, IOracle oracle, SymbolTable symTab,InstanceMapping instanceMapping) {
		int instCount=symTab.registerSimple(reader, options.count, options.labeled);
		if(oracle.canDoNonProj())
			symTab.addSwapTransition();
		symTab.freezeMappings();
		List<Instance> instances=instanceMapping.createInstances(reader,symTab,instCount,oracle,options.labeled);
		return instances;
	}

	public static IOracle sanityCheckProjectivity(Options options,SentenceReader<?> reader, IOracle oracle) {
		int firstNPSent=findNPSent(reader);
		if(oracle.canDoNonProj()){ //if oracle can do nonproj, then we need to have at least one nonproj sent
			if(firstNPSent==-1){
				System.err.println("\n\nTraining data contains NO non-proj sentences -- reverting to ArcStandard\n\n");
				oracle=OracleFactory.getVanillaOracle(options.oracle, TransitionSystem.ArcS);
			}
		} else { //if not, then there can't be any
			if(firstNPSent>=0){
				failCantDoNonProj(firstNPSent,oracle);
			}
		}
		return oracle;
	}

	static SentenceReader<?> getTrainReader(Options options) {
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.train, options.format.sentenceClass);
		if(options.discardNonProj)
			reader=NonProjDiscardingReader.createReader(reader);
		if(options.reversed)
			reader=ReversingReader.createReader(reader);
		return reader;
	}

	public static double getAvgTransSeqLen(List<Instance> instances) {
		double sum=0;
		for(Instance inst:instances)
			sum+=inst.transSequence.length;
		return sum/instances.size();
	}

	public static void failCantDoNonProj(int firstNPSent, IOracle oracle) {
		System.out.println("The oracle "+oracle.getClass().getCanonicalName()+" cannot handle non-projective trees. Checking that your training data is projective.");
		System.out.println("But sent "+firstNPSent+ " is non-projective. Aborting.");
		System.exit(1);
	}

	public static int findNPSent(SentenceReader<?> reader) {
		int i=0;
		for(ISentence s:reader){
			boolean nonProj=!NonProjInfo.isProjectiveSentence(s.getGoldHeads());
			if(nonProj)
				return i;
			++i;
		}
		return -1;
	}
	
	public static List<List<Instance>> cvSplit(List<Instance> instances,int folds){
		List<List<Instance>> r=new ArrayList<List<Instance>>(); 
		if(instances.size()<folds)
			throw new Error("can't cv with fewer instances than folds");
		int br=instances.size()/folds;
		int b=0;
		int n=br;
		for(int i=0;i<folds-1;++i){
			r.add(instances.subList(b, n));
			b=n;
			n+=br;
		}
		r.add(instances.subList(b, instances.size()));
		return r;
	}

}
