package ims.trans.doc;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.DocumentReader;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.perceptron.Long2Int;
import ims.perceptron.Long2IntExact;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.Regularizer;
import ims.trans.AbstractTrainer;
import ims.trans.Decoder;
import ims.trans.FeatureExtractor;
import ims.trans.GreedyTrainer;
import ims.trans.InstanceMapping;
import ims.trans.LossFunction;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.trans.features.FeatureExtractorFactory;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFilterFactory;
import ims.trans.StaticOracleTrainer;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.ZipUtil;

import java.io.IOException;
import java.util.List;
import java.util.zip.ZipOutputStream;

public class Train {

	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		options.warnNoModel();
		
		if(options.testEveryIter && options.out==null){
			System.err.println("-testEvery but no -out given, stopping.");
			System.exit(1);
		}

		IOracle oracle=OracleFactory.getStaticOracle(TransitionSystem.ArcSNewSwap);
		
		SymbolTable symTab=new SymbolTable(options.useStag,true);

		InstanceMapping instanceMapping=new InstanceMapping(options.useStag,options.useStack);
		
		DocumentReader<?> reader=DocumentReader.getReader(options.format.sentenceClass, options.train);
		
		List<Instance> instances = createInstances(options, reader, oracle, symTab,	instanceMapping);
		
		if(Options.shuffleStrategy.shuffleFirstIteration){
			DBO.println("Shuffling instances");
			new ArrayShuffler().shuffle(instances);
		}
		
		double avgTransSeqLen=ims.trans.Train.getAvgTransSeqLen(instances);
		DBO.printlnNoPrefix("Average transition sequence length (canonical):   "+String.format("%8.2f",avgTransSeqLen));

		LossFunction lossFunction=LossFunction.getLossFunction(Options.lossFunctionType);
		Regularizer regularizer=Regularizer.getRegularizer(options.regularizerType, options.C,options.eta,options.lambda);
		Long2IntInterface l2i=options.useHashKernel?new Long2Int(options.hashSize):new Long2IntExact(options.hashSize);		
		ParametersFloat parameters=new ParametersFloat(l2i.size(),regularizer);
		ITransitionFilter transitionFilter=TransitionFilterFactory.getDocumentTransitionFilter(symTab);
		FeatureExtractor featureExtractor=FeatureExtractorFactory.getExtractor(symTab,options);
		Decoder decoder=new Decoder(parameters.parameters, l2i, symTab, transitionFilter,featureExtractor,options.mergeEquivalentStates,options.labeled,instanceMapping);
		if(options.initBias)
			featureExtractor.initWeights(symTab, parameters, l2i);

		AbstractTrainer trainer;
		if(options.trainGreedy){
			options.beam=1;
			trainer=new GreedyTrainer(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping);
		} else {
			trainer=new StaticOracleTrainer(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping, Options.goldBeam);
		}
		Model m=trainer.train();
		DBO.println("Done training, saving model");
		ZipOutputStream zos=ZipUtil.openZOS(options.model);
		ZipUtil.writeObjectAsZipEntry(zos, Model.ZIP_FILE_ENTRY, m);
		zos.close();
		DBO.println("Done.");
		options.done();
	}

	private static List<Instance> createInstances(Options options, DocumentReader<?> reader, IOracle oracle, SymbolTable symTab, InstanceMapping instanceMapping) {
		int instCount=symTab.registerSimple(reader.getSentenceIterator(), options.count, options.labeled);
		symTab.addSwapTransition();
		symTab.addSBTransition();
		List<Instance> l=instanceMapping.createTrainInstances(reader, symTab, oracle, options.labeled);
		symTab.freezeMappings();
		return l;
	}
}
