package ims.trans.transitions;

import java.util.HashMap;
import java.util.Map;

import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.StringSymbolMapping;
import ims.oracle.Configuration.TR;
import ims.trans.State;

public class TransitionFactory {

	public static final String NOLABEL="NOLABEL";
	public static final int NOHEAD_LABEL_L=AbstractSymbolMapping.NONE_IDX;
	public static final int UNKNOWN_LABEL_L=AbstractSymbolMapping.UNK_IDX;
	
	private Map<String,ITransition> cache=new HashMap<String,ITransition>();
	
	public ITransition getTransition(String opLong,int id,StringSymbolMapping deprels){
		ITransition ret=cache.get(opLong);
		if(ret!=null)
			return ret;
		
		String label=NOLABEL;
		String op=opLong;
		int l=UNKNOWN_LABEL_L;
		if(opLong.contains("_")){
			String[] a=op.split("_",2);
			op=a[0];
			label=a[1];
			l=deprels.lookup(label);
		}
		if(op.equals(TR.SH.toString()))
			ret=new Shift(id);
		else if(op.equals(TR.LA.toString()))
			ret=new LeftArc(label,id,l);
		else if(op.equals(TR.RAS.toString()))
			ret=new RightArcS(label,id,l);
		else if(op.equals(TR.SW.toString()))
			ret=new Swap(id);
		else if(op.equals(TR.SB.toString()))
			ret=new SentenceBoundary(id);
		else
			throw new Error("unknown transition: "+op);
		cache.put(opLong, ret);
		return ret;
	}
	
	static abstract class AbstractTransition implements ITransition {
		private static final long serialVersionUID = -6456181295878992224L;
		
		protected final int id;
		AbstractTransition(int id){
			this.id=id;
		}
		
		@Override
		public int getId(){
			return id;
		}
		
		@Override
		public boolean equals(ITransition other) {
			return id==other.getId();
		}
		
		@Override
		public String getBasicTransitionLabel(){
			return getTR().toString();
		}
	}
	
	public static abstract class LabeledAbstractTransition extends AbstractTransition {
		private static final long serialVersionUID = -3606148454912347956L;
		
		protected final String lbl;
		public final int l; //This is the numerical id of a label, as taken from SymbolTable.deprels
		
		private final boolean isLabeled;
		
		LabeledAbstractTransition(int id,String lbl,int l){
			super(id);
			this.lbl=lbl;
			this.l=l;
			this.isLabeled=!lbl.equals(NOLABEL);
		}
		
		@Override
		public String getFullTransitionLabel(){
			String l=getBasicTransitionLabel();
			if(lbl.equals(NOLABEL))
				return l;
			else
				return l+"_"+lbl;
		}
		
		public boolean isLabeled() {
			return isLabeled;
		}
	}
	
	static final class SentenceBoundary extends AbstractTransition{
		private static final long serialVersionUID = 9020093724493633924L;
		
		SentenceBoundary(int id) {
			super(id);
		}

		@Override
		public State apply(State state, double newScore) {
			return state.doSB(this, newScore);
		}

		@Override
		public String getFullTransitionLabel(){
			return getBasicTransitionLabel();
		}

		@Override
		public TR getTR() {
			return TR.SB;
		}

		@Override
		public boolean isLabeled() {
			return false;
		}

		@Override
		public int getDep(State state) {
			return -1;
		}

		@Override
		public boolean isArcTransition() {
			return false;
		}
		
	}
	
	static final class Shift extends AbstractTransition{
		private static final long serialVersionUID = 32985732598732L;
		
		public Shift(int id){
			super(id);
		}
		
		@Override
		public State apply(State state,double newScore) {
			return state.doShift(this, newScore);
		}

		@Override
		public String getFullTransitionLabel(){
			return getBasicTransitionLabel();
		}

		@Override
		public TR getTR() {
			return TR.SH;
		}

		@Override
		public boolean isLabeled() {
			return false;
		}

		@Override
		public int getDep(State state) {
			return -1;
		}
		
		public String toString(){
			return "Shift";
		}

		@Override
		public boolean isArcTransition() {
			return false;
		}

	}
	
	static final class LeftArc extends LabeledAbstractTransition {
		private static final long serialVersionUID = 675323866092047745L;
		
		public LeftArc(String label,int id, int l){
			super(id,label,l);
		}
		
		@Override
		public State apply(State state, double newScore) {
			return state.doLeftArc(this, newScore);
		}
		
		@Override
		public TR getTR() {
			return TR.LA;
		}
		
		@Override
		public int getDep(State state) {
			return state.peekStack(1);
		}
		
		public String toString(){
			return "LeftArc-"+l;
		}

		@Override
		public boolean isArcTransition() {
			return true;
		}

	}
	
	static final class RightArcS extends LabeledAbstractTransition {
		private static final long serialVersionUID = -2664950892449578404L;

		public RightArcS(String label,int id, int l){
			super(id,label,l);
		}
		
		@Override
		public State apply(State state, double newScore) {
			return state.doRightArcS(this, newScore);
		}
		
		@Override
		public TR getTR() {
			return TR.RAS;
		}
		
		@Override
		public int getDep(State state) {
			return state.peekStack(0);
		}
		
		public String toString(){
			return "RightArc-"+l;
		}

		@Override
		public boolean isArcTransition() {
			return true;
		}

	}
	
	static final class Swap extends AbstractTransition {
		private static final long serialVersionUID = -4415573022346544175L;

		public Swap(int id) {
			super(id);
		}

		@Override
		public State apply(State state, double newScore) {
			return state.doSwap(this, newScore);
		}

		@Override
		public String getFullTransitionLabel() {
			return getBasicTransitionLabel();
		}
		
		@Override
		public TR getTR() {
			return TR.SW;
		}
		
		@Override
		public boolean isLabeled() {
			return false;
		}

		@Override
		public int getDep(State state) {
			return -1;
		}
		
		public String toString(){
			return "Swap";
		}

		@Override
		public boolean isArcTransition() {
			return false;
		}

	}
}
