package ims.trans.util;

import java.io.File;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import ims.data.ISentence;
import ims.io.Format;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.trans.Options;
import ims.util.Accuracy;
import ims.util.MutableInt;
import ims.util.Pair;
import ims.util.PrecisionRecall;

public class Eval {
	
	public static Pair<Accuracy,Accuracy> getUASandLAS(File pred,File gold,Format format){
		return getUASandLAS(pred,gold,format.sentenceClass);
	}
	
	public static <T extends ISentence> Pair<Accuracy,Accuracy> getUASandLAS(File pred,File gold,Class<T> sentenceClass){
		SentenceReader<T> predReader=ReaderWriterFactory.getReader(pred, sentenceClass);
		SentenceReader<T> goldReader=ReaderWriterFactory.getReader(gold, sentenceClass);
		return getUASandLAS(goldReader,predReader);
	}
	
	public static <T extends ISentence> Pair<Accuracy,Accuracy> getUASandLAS(SentenceReader<T> gold,SentenceReader<T> pred){
		int total=0;
		int corrL=0;
		int corrU=0;
		
		for(Iterator<T> gIt=gold.iterator(),pIt=pred.iterator();gIt.hasNext();){
			ISentence g=gIt.next();
			ISentence p=pIt.next();
			if(!sameWords(g,p))
				throw new Error("not same words while evaluating");
			int[] gHeads=g.getGoldHeads();
			int[] pHeads=p.getPredHeads();
			String[] gLbl=g.getGoldLabels();
			String[] pLbl=p.getPredLabels();
			for(int i=1;i<gHeads.length;++i){
				++total;
				if(gHeads[i]==pHeads[i]){
					corrU++;
					if(gLbl[i].equals(pLbl[i]))
						corrL++;
				}
			}
		}
		
		return new Pair<Accuracy,Accuracy>(new Accuracy(total,corrU),new Accuracy(total,corrL));
	}

	static boolean sameWords(ISentence g, ISentence p) {
		String[] gForms=g.getForms();
		String[] pForms=p.getForms();
		if(gForms.length!=pForms.length)
			return false;
		for(int i=1;i<gForms.length;++i){
			if(!gForms[i].equals(pForms[i]))
				return false;
		}
		return true;
	}
	
	public static void main(String[] args){
		Options options=new Options(args);
		
		Format goldFormat=options.format;
		Format predFormat=options.format;
		if(options.goldFormat!=null)
			goldFormat=options.goldFormat;
		if(options.predFormat!=null)
			predFormat=options.predFormat;
		
		SentenceReader<?> pred=ReaderWriterFactory.getReader(options.out, predFormat);
		SentenceReader<?> gold=ReaderWriterFactory.getReader(options.gold, goldFormat);
		
		int total=0;
		int corrL=0;
		int corrU=0;

		
		Map<String,MutableInt> lblTP=new TreeMap<String,MutableInt>();
		Map<String,MutableInt> lblFN=new TreeMap<String,MutableInt>();
		Map<String,MutableInt> lblFP=new TreeMap<String,MutableInt>();
		
		Map<String,MutableInt> totalByPos=new TreeMap<String,MutableInt>();
		Map<String,MutableInt> correctLByPos=new TreeMap<String,MutableInt>();
		
		for(Iterator<? extends ISentence> gIt=gold.iterator(),pIt=pred.iterator();gIt.hasNext();){
			ISentence g=gIt.next();
			ISentence p=pIt.next();
			if(!Eval.sameWords(g,p))
				throw new Error("not same words while evaluating");
			int[] goldHeads=g.getGoldHeads();
			int[] predHeads=p.getPredHeads();
			String[] goldLbl=g.getGoldLabels();
			String[] predLbl=p.getPredLabels();
			
			String[] gPos=g.getGoldPOS();
			
			for(int i=1;i<goldHeads.length;++i){
				++total;
				boolean uCorrect=goldHeads[i]==predHeads[i];
				boolean lCorrect=goldHeads[i]==predHeads[i] && goldLbl[i].equals(predLbl[i]);
				
				String pos=gPos[i];
				MutableInt.incrementMapValue(totalByPos, pos);
				
				String gLbl=goldLbl[i];
				String pLbl=predLbl[i];
				
				if(uCorrect)
					++corrU;
				if(lCorrect){
					++corrL;
					MutableInt.incrementMapValue(correctLByPos, pos);
					MutableInt.incrementMapValue(lblTP, gLbl);
				} else {
					MutableInt.incrementMapValue(lblFN, gLbl);
					MutableInt.incrementMapValue(lblFP, pLbl);
				}
			}
		}
		
		//Now print all the stuff out
		Accuracy UAS=new Accuracy(total,corrU);
		Accuracy LAS=new Accuracy(total,corrL);
		
		System.out.println("UAS:  "+UAS.toString());
		System.out.println("LAS:  "+LAS.toString());
		
		System.out.println();
		System.out.println();
		System.out.println("===== Labeled Accuracy by dependent POS =====");
		for(String tag:totalByPos.keySet()){
			Accuracy acc=new Accuracy(totalByPos.get(tag),correctLByPos.get(tag));
			System.out.printf("%-10s   %s\n",tag,acc.toString());
		}
		
		System.out.println();
		System.out.println();
		System.out.println("===== P/R/F by label =====");
		
		Set<String> labels=new TreeSet<String>();
		labels.addAll(lblTP.keySet());
		labels.addAll(lblFP.keySet());
		labels.addAll(lblFN.keySet());
		for(String lbl:labels){
			MutableInt tp=lblTP.get(lbl);
			MutableInt fp=lblFP.get(lbl);
			MutableInt fn=lblFN.get(lbl);
			PrecisionRecall pr=new PrecisionRecall(tp,fp,fn);
			System.out.printf("%-12s   %s  |  %s  |  %s\n",lbl,pr.getPrecisionStr(),pr.getRecallStr(),pr.getF1Str());
		}
	}
	
}
