package ims.util;

import java.io.IOException;
import java.util.Iterator;

/**
 * Standard way to implement the iterator for something that implements 
 * Iterable<T> for Reader kind of things.
 *  
 * 
 * @author anders
 *
 * @param <T> type to iterate over.
 */
public abstract class AbstractSingleItemBufferedIterator<T> implements Iterator<T> {

	private T next;
	
	//XXX Make sure to have a call of the form super.initSetNext(init()) in the constructor of any subclass
	
	protected AbstractSingleItemBufferedIterator() {}
	
	protected abstract T readNext() throws IOException;
	protected abstract T init() throws IOException;
	
	/**
	 * The annoying thing about this design pattern is that reading the very first 
	 * item may be very implementation specific. 
	 * 
	 * Since subclass constructors start by calling their superconstructors, it is 
	 * difficult to delegate to the superconstructor to buffer up the first item.
	 * 
	 * An alternative would be to leave the field 'next' protected, and assume that
	 * the constructor of the subclass handles this. This is essentially equivalent
	 * to the pattern prescribed here, where every subclass is *supposed* to call
	 * initSetNext() before they exit their constructors. The main difference is that
	 * that particular field will be left accessible to subclasses altogether, which
	 * is probably a bad idea.
	 * 
	 * HENCE,
	 * Subclasses of this class are expected to call this method before exiting the 
	 * constructor. And never again.
	 * 
	 */
	protected void initSetNext(T n){
		this.next=n;
	}
	@Override
	public final boolean hasNext() {
		return next!=null;
	}
	
	@Override
	public final T next() {
		T r=next;
		try {
			next=readNext();
		} catch(IOException e){
			throw new RuntimeException("Error while reading next item",e);
		}
		return r;
	}

	@Override
	public final void remove() {
		throw new UnsupportedOperationException("Remove not supported");
	}

	public final T peekNext(){
		return next;
	}
}
