package ims.util;

import java.util.ArrayList;
import java.util.List;

import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.AbstractDynamicOracle;
import ims.oracle.AbstractStaticOracle;
import ims.oracle.MinimalConfiguration;
import ims.oracle.OracleFactory;
import ims.oracle.Configuration.TR;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;

public class DataSetStats {

	/*
	 * total number of sentences
	 */
	static int sentCount=0;
	
	/*
	 * total number of tokens
	 */
	static int tokenCount=0;
	
	/*
	 * senLengthCounts[i] denotes the number of sentences of length i
	 */
	static int[] senLengthCounts=new int[1000];
	
	/*
	 * nonProjArcSenCounts[i] denotes the number of sentences with i non-projective arcs 
	 */
	static int[] nonProjArcSenCounts=new int[50];
	
	/*
	 * Total number of non-projective arcs
	 */
	static int npArcCount=0;
	
	/*
	 * trSenCounts[a][i] denotes the number of sentences that need i transitions for oracles
	 * a=0 -- old swap
	 * a=1 -- new swap
	 * a=2 -- optimal swap
	 */
	static int[][] trSenCounts=new int[3][2000];
	
	/*
	 * Total number of transitions
	 */
	static int[] trSum=new int[3];
	
	/*
	 * Total number of swaps
	 */
	static int[] swSum=new int[3];

	/*
	 * How often are the derivations strictly unique?
	 */
	static int sentNoAmibguity=0;
	
	/*
	 * Same as above, but only projective
	 */
	static int projSentNoAmbiguity=0;
	
	/*
	 * On how many sentences do the oracles coincide?
	 */
	static int allOraclesCoincide=0;
	static int oldNewCoincide=0;
	static int newOptCoincide=0;
	static int oldOptCoincide=0;
	
	private static void doCounts(SentenceReader<?> reader) {
		AbstractStaticOracle[] oracles=new AbstractStaticOracle[3];
		oracles[0]=OracleFactory.getStaticOracle(TransitionSystem.ArcSOldSwap);
		oracles[1]=OracleFactory.getStaticOracle(TransitionSystem.ArcSNewSwap);
		oracles[2]=OracleFactory.getStaticOracle(TransitionSystem.ArcSOptimalSwap);
		for(ISentence s:reader){
			int[] heads=s.getTrainHeads();
			int length=heads.length-1;
			++sentCount;
			tokenCount+=length;
			senLengthCounts[length]++;
			
			//Non-proj
			NonProjInfo npi=new NonProjInfo(heads);
			nonProjArcSenCounts[npi.totalNonProjArcs]++;
			npArcCount+=npi.totalNonProjArcs;
			
			//Oracles and transitions
			List<List<String>> seqs=new ArrayList<List<String>>();
			for(int i=0;i<3;++i){
				oracles[i].initSentence(heads);
				List<String> trs=oracles[i].getCanonicalSequence(heads, null, false);
				seqs.add(trs);
				trSenCounts[i][trs.size()]++;
				trSum[i]+=trs.size();
				swSum[i]+=countSwaps(trs);
			}
			boolean oldNewEQ=seqs.get(0).equals(seqs.get(1));
			boolean oldOptEQ=seqs.get(0).equals(seqs.get(2));
			boolean newOptEQ=seqs.get(1).equals(seqs.get(2));
			if(oldNewEQ)
				oldNewCoincide++;
			if(oldOptEQ)
				oldOptCoincide++;
			if(newOptEQ)
				newOptCoincide++;
			
			if(oldNewEQ && newOptEQ)
				allOraclesCoincide++;
			
			boolean sentenceHasUniqueTrSeq=hasUniqueTrSeq(heads);
			if(sentenceHasUniqueTrSeq){
				++sentNoAmibguity;
				if(!npi.nonProjectiveSentence)
					++projSentNoAmbiguity;
			}
		}
	}

	private static final AbstractDynamicOracle dynSwapOracle=OracleFactory.getDynamicOracle(TransitionSystem.ArcSDynSwap);
	public static boolean hasUniqueTrSeqNoArcAmbig(int[] heads) {
		dynSwapOracle.initSentence(heads);
		MinimalConfiguration conf=new MinimalConfiguration(heads);
		while(!conf.isTerminal()){
			List<TR> trs=dynSwapOracle.getActions(conf, heads);
			if(trs.size()>1)
				return false;
			conf.applyTR(trs.get(0));
		}
		return true;
	}
	
	private static final AbstractDynamicOracle fullDynSwapOracle=OracleFactory.getDynamicOracle(TransitionSystem.ArcSFullDynSwap);
	
	public static boolean hasUniqueTrSeq(int[] heads) {
		fullDynSwapOracle.initSentence(heads);
		MinimalConfiguration conf=new MinimalConfiguration(heads);
		while(!conf.isTerminal()){
			List<TR> trs=fullDynSwapOracle.getActions(conf, heads);
			if(trs.size()>1)
				return false;
			conf.applyTR(trs.get(0));
		}
		return true;
	}


	private static int countSwaps(List<String> trs) {
		int c=0;
		for(String s:trs)
			if(s.equals(TR.SW.toString()))
				++c;
		return c;
	}
	
	private static void printStats() {
		System.out.println("===== Overall statistics =====");
		System.out.printf("Total sentences:   %,8d\n",sentCount);
		System.out.printf("Projective:        %,8d   -- %6.2f%%\n",nonProjArcSenCounts[0],100.0*nonProjArcSenCounts[0]/sentCount);
		System.out.printf("Non-projective:    %,8d   -- %6.2f%%\n",sentCount-nonProjArcSenCounts[0],100.0*(sentCount-nonProjArcSenCounts[0])/sentCount);
		System.out.println();
		System.out.printf("Sentence w/o amb:  %,8d   -- %6.2f%%\n",sentNoAmibguity,100.0*sentNoAmibguity/sentCount);
		System.out.printf("--projective:      %,8d   -- %6.2f%%\n",projSentNoAmbiguity,100.0*projSentNoAmbiguity/sentCount);
		System.out.println();
		System.out.printf("Total tokens:      %,8d\n",tokenCount);
		System.out.printf("Average tokens:    %8.2f\n",(double)tokenCount/sentCount);
		System.out.println();
		System.out.println();
		System.out.println("===== Number of sentences with i non-projective arcs ==== ");
		System.out.println("i          count       %         acc%");
		System.out.println();
		int acc=0;
		for(int i=0;i<nonProjArcSenCounts.length;++i){
			if(nonProjArcSenCounts[i]==0)
				continue;
			double ratio=100.0*nonProjArcSenCounts[i]/sentCount;
			acc+=nonProjArcSenCounts[i];
			double accRatio=100.0*acc/sentCount;
			System.out.printf("%-3d     %,8d    %6.2f%%    %6.2f%%\n",i,nonProjArcSenCounts[i],ratio,accRatio);
		}
		System.out.println();
		System.out.println();
		System.out.println("===== Transition statistics =====");
		System.out.println("Total transitions:");
		System.out.printf("--SwapOld:  %,8d\n",trSum[0]);
		System.out.printf("--SwapNew:  %,8d\n",trSum[1]);
		System.out.printf("--SwapOpt:  %,8d\n",trSum[2]);
		System.out.println();
		System.out.println("Avg transitions per sentence:");
		System.out.printf("--SwapOld:  %.2f\n",(double)trSum[0]/sentCount);
		System.out.printf("--SwapNew:  %.2f\n",(double)trSum[1]/sentCount);
		System.out.printf("--SwapOpt:  %.2f\n",(double)trSum[2]/sentCount);
		System.out.println();
		System.out.println("Total number of swaps and reduction relative to Old:");
		System.out.printf("--SwapOld:  %,8d   %6.2f%%\n",swSum[0],100.0*(swSum[0]-swSum[0])/swSum[0]);
		System.out.printf("--SwapNew:  %,8d   %6.2f%%\n",swSum[1],100.0*(swSum[0]-swSum[1])/swSum[0]);
		System.out.printf("--SwapOpt:  %,8d   %6.2f%%\n",swSum[2],100.0*(swSum[0]-swSum[2])/swSum[0]);
		System.out.println();
		System.out.println();
		
		System.out.println("===== Oracles coincide =====");
		System.out.printf("--Old-New:    %,8d  %6.2f%%\n",oldNewCoincide,100.0*oldNewCoincide/sentCount);
		System.out.printf("--New-Opt:    %,8d  %6.2f%%\n",newOptCoincide,100.0*newOptCoincide/sentCount);
		System.out.printf("--Old-Opt:    %,8d  %6.2f%%\n",oldOptCoincide,100.0*oldOptCoincide/sentCount);
		System.out.printf("--ALL:        %,8d  %6.2f%%\n",allOraclesCoincide,100.0*allOraclesCoincide/sentCount);
		System.out.println();
		System.out.println();
		System.out.println("===== Sentence length histogram (with acc) =====");
		int accS=0;
		for(int i=0;i<senLengthCounts.length;++i){
			if(senLengthCounts[i]==0)
				continue;
			double ratio=100.0*senLengthCounts[i]/sentCount;
			accS+=senLengthCounts[i];
			double accRatio=100.0*accS/sentCount;
			System.out.printf("%-5d     %,8d     %6.2f%%     %6.2f%%\n",i,senLengthCounts[i],ratio,accRatio);
		}
		
	}


	public static void main(String[] args){
		Options options=new Options(args);
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.train,options.format.sentenceClass);
		doCounts(reader);
		printStats();	
	}

}
