package ims.util;

import java.util.Random;

/*
 * The basic idea here is that we have reproducible sequences of random doubles that obeys two levels of rolling back.
 * 
 * In the context of beam search over sentences with states this boils down to:
 * - for every sentence in an iteration we are able to rollback
 * - for every transition within a state in a sentence we are able to rollback
 * - for every feature index ix within a state and sentence we want the same double back
 * 
 * This basically boils down to three indices:
 * r(sent,state,tr,ix)
 * 
 * Thus, given x, state, and ix, we want to get the same double back.
 * 
 * r(x,ind(state),t_?,ix) == r(x,ind(state),t_?,ix)
 * 
 * And, since we never need to rollback x, the unique coordinates of a double is described by ind(state) and ix
 *
 * 
 * We implement this by the following:
 * - When we start a new sentence, rDeep is reseeded with a new seed from deepestRNG
 * - When we move from one sentence from the next, we advance the deepestRNG, and use this for the rDeep/rbDeepSeed
 * - When we advance from one state to the next, we advance rDeep RNG, and use this for rbShallowSeed/rShallow
 */

public class LeveledRollbackRNG {

	long stateSeed;
	long featureSeed;
	long lastFSeed;

	Random deepestRNG;
	Random stateRandom;
	Random featureRandom;
	
	public LeveledRollbackRNG(long seed){
		deepestRNG=new Random(seed);
		stateSeed=deepestRNG.nextLong();
		stateRandom=new Random(stateSeed);
		featureSeed=stateRandom.nextLong();
		featureRandom=new Random(featureSeed);
	}
	
	public void rollbackSentence(){
		stateRandom.setSeed(stateSeed);           //Reset deep seed
		featureSeed=stateRandom.nextLong();
		featureRandom.setSeed(featureSeed);
	}
	
	public void rollbackFeatures(){
		featureRandom.setSeed(featureSeed);     //Rollback shallow RNG
	}
	
	public void advanceSentence(){
		stateSeed=deepestRNG.nextLong();
		rollbackSentence();
	}
	
	public void advanceState(){
		long fSeed=stateRandom.nextLong();        //Get shallow seed
		lastFSeed=featureSeed;
		featureSeed=fSeed;                        //Store shallow seed
		featureRandom.setSeed(fSeed);             //Set shallow see
	}
	
	public double nextDouble(){
		return featureRandom.nextDouble();
	}
	
	public void revertState() {
		featureSeed=lastFSeed;
		rollbackFeatures();
	}
	
	public static void main(String[] args){
		LeveledRollbackRNG a=new LeveledRollbackRNG(123);
		for(int sent=0;sent<5;++sent){
			System.out.println("\n\n\n ==== Sentence "+sent+" ==== \n\n");
			for(int i=0;i<2;++i){
				System.out.println("ROUND "+i+"\n\n\n");
				for(int state=0;state<3;++state){
					System.out.println("State "+state);
					for(int tr=0;tr<3;++tr){
						System.out.println("Trans "+tr);
						for(int f=0;f<3;++f){
							System.out.println("Featue f"+f+": "+a.nextDouble());
						}
						a.rollbackFeatures();
					}
					a.advanceState();
				}
				a.rollbackSentence();
			}
			a.advanceSentence();
		}
		
	}

}
