package ims.util;

import ims.data.ISentence;

import java.util.Arrays;

public class NonProjInfo {

	private static final int[] EMPTY_INT_ARR=new int[0];
	private static final String EMPTY_LEVEL_SIG="<>";
	
	public final int rootDeps;
	public final int[] depth;
	public final boolean[][] ancestors;
	
	public final boolean nonProjectiveSentence;
	public final boolean illNestedSentence;
	
	
	public final boolean[] isNonProjectiveEdge;
	public final int[][] gaps;
	public final int[] intervalDegree;     //Interval Degree is Havelka's term. Kuhlmann and Nivre (2006) call (the max of) this gap degree 
	public final int[] levelType;
	public final int[][] levelSignatures;
	public final int[][] componentRoots;   //These are the roots of the components, i.e., componentRoots[dep] is an array of all the root components

	public final int totalNonProjArcs;
	
	public NonProjInfo(int[] heads){
		rootDeps=countRootDeps(heads);
		ancestors=getAncestors(heads);
		depth=getDepth(heads);
		
		int npArc=0;
		isNonProjectiveEdge=new boolean[heads.length];
		for(int i=1;i<heads.length;++i)
			if(!isProjective(i,heads)){
				isNonProjectiveEdge[i]=true;
				npArc++;
			}
		totalNonProjArcs=npArc;
		boolean proj=npArc==0;

		intervalDegree=new int[heads.length];
		levelType=new int[heads.length];
		levelSignatures=new int[heads.length][];
		componentRoots=new int[heads.length][];
		if(proj){ //The sentence is projective, pretty boring.
			nonProjectiveSentence=false;
			illNestedSentence=false;
			gaps=new int[heads.length][];
			for(int i=0;i<heads.length;++i){
				levelSignatures[i]=EMPTY_INT_ARR;
				componentRoots[i]=EMPTY_INT_ARR;
				gaps[i]=EMPTY_INT_ARR;
			}
		} else { //Compute other stuff
			gaps=getGapSets(heads);
			nonProjectiveSentence=true;
			illNestedSentence=!wellNested(gaps,heads);
			for(int dep=0;dep<heads.length;++dep){
				intervalDegree[dep]=getIntervalDegree(gaps[dep]);
				levelType[dep]=getLevelType(dep,depth,gaps[dep]);
				componentRoots[dep]=getComponentRoots(gaps[dep],heads,dep);
				levelSignatures[dep]=getLevelSignature(componentRoots[dep],depth,dep);
			}
		}
	}
	
	public static boolean[] getNonProjArcs(int[] heads) {
		boolean[] isNonProjectiveEdge=new boolean[heads.length];
		for(int i=1;i<heads.length;++i)
			if(!isProjective(i,heads))
				isNonProjectiveEdge[i]=true;
		return isNonProjectiveEdge;
	}

	public void sanityCheck(){ //Test some conditions that should hold
		//a non-positive level-signature entails illnestedness:
		OUTER:
		for(int a[]:levelSignatures)
			for(int b:a)
				if(b<=0){
					if(!illNestedSentence)
						throw new Error("should be illnested if there is nonpositive level sig");
					break OUTER;
				}
		for(int dep=0;dep<depth.length;++dep){
			if(isNonProjectiveEdge[dep]!=(gaps[dep].length!=0))
				throw new Error("Disagreement between non-proj edge and gap set");
			if(isNonProjectiveEdge[dep])
				if(levelType[dep]!=max(levelSignatures[dep]))
					throw new Error("Disagreement between level type and level signature");
			if(componentRoots[dep].length!=levelSignatures[dep].length)
				throw new Error("disagreement between number of component roots and size of level sig");
		}
		
	}
	
	public static boolean isProjectiveSentence(int[] heads){
		for(int i=1;i<heads.length;++i){
			if(!isProjective(i,heads))
				return false;
		}
		return true;
	}

	public static boolean isProjective(int dep,int[] heads){
		int head=heads[dep];
		if(head==0)
			return true;
		int min=dep;
		int max=head;
		if(head<dep){
			min=head;
			max=dep;
		}
		for(int i=min+1;i<max;++i){
			int cur=i;
			while(cur!=dep && cur!=head){
				if(cur==0)
					return false;
				cur=heads[cur];
			}
		}
		return true;
	}

	public static int[][] getGapSets(int[] heads){
		int[][] r=new int[heads.length][];
		r[0]=EMPTY_INT_ARR;
		for(int dep=1;dep<heads.length;++dep){
			int head=heads[dep];
			int min=Math.min(dep, head);
			int max=Math.max(dep, head);
			int cnt=0;
			for(int j=min+1;j<max;++j)
				if(!isDominatedBy(j,head,heads))
					cnt++;
			if(cnt>0){
				int[] p=new int[cnt];
				int c=0;
				for(int j=min+1;j<max;++j)
					if(!isDominatedBy(j,head,heads))
						p[c++]=j;
				r[dep]=p;
			} else {
				r[dep]=EMPTY_INT_ARR;
			}
		}
		return r;
	}

	private static boolean isDominatedBy(int node,int dominator,int[] heads) {
		for(int cur=node;cur!=ISentence.ROOT_HEAD;cur=heads[cur])
			if(cur==dominator)
				return true;
		return false;
	}
	
	public static boolean wellNested(int[][] gapSets,int[] heads){
		for(int i=1;i<heads.length;++i){
//			if(!nonProjEdges[i])
			if(gapSets[i].length==0)
				continue;
			int j=heads[i];
			//The edge i<->j is np. Now see if there is any edge a p<->q such that:
			// - p or q is in the gap of i<->j (== gapSets[i]), and
			// - i or j is in the gap of p<->q (== gapSets[p])
			int[] gap1=gapSets[i];
			for(int p=1;p<heads.length;++p){
				if(p==i || gapSets[p].length==0)
					continue;
				int q=heads[p];
				int[] gap2=gapSets[p];
				boolean foundInGap1=false;
				boolean foundInGap2=false;
				for(int a:gap1)
					if(a==p || a==q){
						foundInGap1=true;
						break;
					}
				for(int a:gap2){
					if(a==i || a==j){
						foundInGap2=true;
						break;
					}
				}
				if(foundInGap1 && foundInGap2)
					return false;
			}
		}
		return true;
	}
	
	public static int getIntervalDegree(int[] gapSet){
		if(gapSet.length==0)
			return 0;
		int degree=1;
		for(int i=0;i<gapSet.length-1;++i)
			if(gapSet[i]+1!=gapSet[i+1])
				degree++;
		return degree;
	}
	
	public static int[] getComponentRoots(int[] gap,int[] heads,int dep){
		if(gap.length==0)
			return EMPTY_INT_ARR;
		if(gap.length==1)
			return new int[]{gap[0]};
		
		int head=heads[dep];
		int min=Math.min(dep,head);
		int max=Math.max(dep,head);

		boolean[] s=new boolean[heads.length];
		int rootCount=0;
		for(int g:gap){
			int r=g;
			while(heads[r]>min && heads[r]<max)
				r=heads[r];
			if(!s[r]){
				s[r]=true;
				++rootCount;
			}
		}
		
		int[] cRoots=new int[rootCount];
		int c=0;
		for(int g:gap){
			int r=g;
			while(heads[r]>min && heads[r]<max)
				r=heads[r];
			if(s[r]){
				s[r]=false;
				cRoots[c++]=r;
			}
		}
		return cRoots;
	}
	
	public static int[] getDepth(int[] heads){
		int[] r=new int[heads.length];
		for(int dep=1;dep<r.length;++dep)
			r[dep]=rootDist(dep,heads);
		return r;
	}
	
	public  static int rootDist(int node,int[] heads){
		if(node==0)
			return 0;
		else
			return 1+rootDist(heads[node],heads);
	}

	public static int getLevelType(int node,int[] depth,int[] gap){
		if(gap.length==0)
			return depth[node];
		int minGapLevel=depth[gap[0]];
		for(int p=1;p<gap.length;++p)
			minGapLevel=Math.min(minGapLevel, depth[gap[p]]);
		
		return depth[node]-minGapLevel;
	}

	public static int[] getLevelSignature(int[] cRoots,int[] depth,int dep){
		int[] sig=new int[cRoots.length];
		for(int i=0;i<cRoots.length;++i){
			sig[i]=depth[dep]-depth[cRoots[i]];
		}
		Arrays.sort(sig);
		return sig;
	}
	
	public static String levelSig2String(int[] sig){
		if(sig.length==0)
			return EMPTY_LEVEL_SIG;
		StringBuilder sb=new StringBuilder("<").append(sig[0]);
		for(int i=1;i<sig.length;++i)
			sb.append(",").append(sig[i]);
		return sb.append(">").toString();
	}

	public static int countRootDeps(int[] heads){
		int count=0;
		for(int i=1;i<heads.length;++i){
			if(heads[i]==0)
				count++;
		}
		return count;
	}

	//This array is [ancestor][descendant]
	public static boolean[][] getAncestors(int[] heads) {
		boolean[][] isChild = new boolean[heads.length][heads.length];
		for(int i = 1, l1=1; i < heads.length; i++,l1=i)  
			while((l1= heads[l1]) != ISentence.ROOT_HEAD)
				isChild[l1][i] = true; //i is a descendant of l1
		return isChild;
	}
	
	static int max(int[] a){
		int r=a[0];
		for(int i=1;i<a.length;++i)
			r=Math.max(r, a[i]);
		return r;
	}

}
